/**************************************************************************
 *
 *          Copyright (c) 2002, SeeBeyond Technology Corporation,
 *          All Rights Reserved
 *
 *          This program, and all the routines referenced herein,
 *          are the proprietary properties and trade secrets of
 *          SEEBEYOND TECHNOLOGY CORPORATION.
 *
 *          Except as provided for by license agreement, this
 *          program shall not be duplicated, used, or disclosed
 *          without  written consent signed by an officer of
 *          SEEBEYOND TECHNOLOGY CORPORATION.
 *
 ***************************************************************************/
package com.stc.connector.httpadapter.http;

import java.io.ByteArrayOutputStream;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.util.Vector;
import sun.misc.BASE64Encoder;
import com.stc.connector.httpadapter.util.LogUtil;
import org.apache.log4j.Logger;

/**
 * HTTP client agent.
 *
 * @author $Author: pchan $
 * @version $Revision: 1.17.18.4.2.2 $
 */
public class HttpClientAPI {
    /** Content-type header */
    public static final String CONTENT_TYPE = "Content-Type";
    /** Content-type text value */
    public static final String CONTENT_TYPE_TEXT = "text";
    /** Content-type urlencoded value */
    public static final String CONTENT_TYPE_URLENCODED
        = "application/x-www-form-urlencoded";
    /** Content-length header */
    public static final String CONTENT_LENGTH = "Content-Length";
    /** Accept-type header */
    public static final String ACCEPT_TYPE = "Accept-Type";
    private static final String META_HTTP_EQUIV_REFRESH
        = "<metahttp-equiv=\"refresh\"";

    // QAI 90061 -- Digest Authentication
    public static final int MAX_CHALLENGE_RETRY = 5; // TODO: MAX_CHALLENGE_RETRY could be configurable but larger or equal to 1
    public static final int WWW_AUTHENTICATE_DIGEST_MIN_FIELD_COUNT = 2;
    public static final int WWW_AUTHENTICATE_BASIC_MIN_FIELD_COUNT = 1;
    //
    private int challenge_retry = 0;
    private int nonceCounter = 1;
    //
    // TODO: this might be too generic, may be version info should be included
    // or perhaps a more dynamic cnonce generation logic is required
    public static final String CLIENT_NONCE = "HTTPeWay"; 
    //
    public static final int UNAUTHORIZED_RESPONSE_CODE = 401;
    public static final int SUCCESS_RESPONSE_CODE = 200;
    //
    public static final String WWW_AUTHENTICATE = "WWW-Authenticate";
    public static final String BASIC = "Basic";
    public static final String DIGEST = "Digest";
    public static final String NONCE = "nonce";
    public static final String OPAQUE = "opaque";
    public static final String REALM = "realm";
    public static final String DOMAIN = "domain";
    public static final String STALE = "stale";
    public static final String ALGORITHM = "algorithm";
    public static final String QOP = "qop";
    public static final String USERNAME = "username";
    public static final String CNONCE = "cnonce";
    public static final String NONCECOUNT = "nc";
    public static final String METHOD = "method";
    public static final String URI = "uri";
    //
    private String digest_auth_nonce = "";
    private String digest_auth_nonce_last = "";
    private String digest_auth_opaque = "";
    private String digest_auth_realm = "";
    private String digest_auth_domain = "";
    private String digest_auth_stale = "";
    private String digest_auth_algorithm = "";
    private String digest_auth_qop = "";
    private String digest_auth_qop_filtered = "";
    private String auth_server_challenge = "";
    //
    public static final int BASIC_AUTH = 0;
    public static final int DIGEST_AUTH = 1;
    public static final int AUTO_AUTH = 999;
    //
    // default to basic so that preemptive authentication
    // header can be sent if the challenge happens to be a basic one;
    // if the challenge is digest, then the detection and retry logic
    // will take care of the rest.
    //
    private int authSchemeToAttempt = BASIC_AUTH; 
    private int authScheme = AUTO_AUTH; // default to automatic detection

    // Data type of data that is sent with post or get
    private final String binaryOutgoing = "Binary"; // Sending binary data
    private final String textOutgoing = "Text"; // Sending text data
    private final String formDataOutgoing = "FormData"; // Sending form data
    private final String urlOutgoing = "URL"; // No data,
    private final String requestGet = "GET";
    private final String requestPost = "POST";
    private final String requestHead = "HEAD";
    private final String requestPut = "PUT";
    private final String requestOptions = "OPTIONS";
    private final String requestDelete = "DELETE";
    private final String requestTrace = "TRACE";

    // Http and Https proxy system properties
    private final String httpProxySetProperty = "http.proxySet";
    private final String httpProxyHostProperty = "http.proxyHost";
    private final String httpProxyPortProperty = "http.proxyPort";
    private final String httpsProxySetProperty = "https.proxySet";
    private final String httpsProxyHostProperty = "https.proxyHost";
    private final String httpsProxyPortProperty = "https.proxyPort";

    // Attributes
    private String urlString = null;
    private URL httpURL = null;
    private HashMap headerCache;
    private Vector contentTypeValsCache = null;
    private boolean allowCookie = false;
    private QueryString query = null;
    private String textData = null;
    private byte[] binaryData = null;
    private String outgoingType;
    private HttpAuthenticator httpAuthenticator = null;
    private String proxyUserName = null;
    private String proxyPassWord = null;
    private String httpUserName = null;
    private String httpPassWord = null;
    private String encoding = null;
    private boolean posting = false;
    private String savedURLOnRedirect = null;
    private boolean noHostVerification = false;
    private SSLContextJDKPlatform sslContext = null;
    private HttpClientAPIJDKPlatform httpJDKPlatform = null;
    private HttpURLConnection httpUC = null;

    protected Logger logger = Logger.getLogger(LogUtil.HTTPCLIENT_CATEGORY +
                            getClass().getName());

    private final static int BUFF_SIZE = 4096;
    private boolean beingRedirected = false;
    private int responseCode; // QAI 90061 -- Digest Authentication

    // QAI 95669: indirect consequence of QAI 93942
    private final String HTTP_URL_WARNING_MESSAGE1 = "Please provide a valid non-empty http url";
    private final String HTTP_URL_WARNING_MESSAGE2 = HTTP_URL_WARNING_MESSAGE1 + 
                                                      " either through the default configuration or invoking the setURL method";

    /**
     * Constructs an HttpClientAPI
     *
     * @throws Exception DOCUMENT ME!
     */
    public HttpClientAPI()
        throws Exception {
        // Initialize Header cache and Content-Type values cache
        headerCache = new HashMap();
        contentTypeValsCache = new Vector();
        outgoingType = new String(urlOutgoing);
        encoding = "ASCII";

        if (isJDK13only()) {
            if (this.logger.isDebugEnabled() == true) {
                this.logger.debug("Running in 1.3 JVM");
            }
            httpJDKPlatform = (HttpClientAPIJDKPlatform) (Class.forName(
                    "com." + "stc." + "connector." + "httpadapter." + "http."
                    + "jdk13." + "HttpClientAPIJDK13")).newInstance();
        } else {
            if (this.logger.isDebugEnabled() == true) {
                this.logger.debug("NOT running in 1.3 JVM");
            }
            httpJDKPlatform = (HttpClientAPIJDKPlatform) (Class.forName(
                    "com." + "stc." + "connector." + "httpadapter." + "http."
                    + "jdk14." + "HttpClientAPIJDK14")).newInstance();
        }
    }

    /**
     * Constructs an HttpClientAPI
     *
     * @param urlString Samples -  http://www.somehost.com
     *        http://www.somehost.com/cgi-bin/query
     *        http://www.somehost.com/cgi-bin/query?name1=value1&name2=value2
     *        The name value pairs supplied must be urlencoded. The
     *        QueryString class is preferred when doing a Post or Get with
     *        mutlitple name value pairs rather than hardcoded in the URL as
     *        in the third example.
     *
     * @throws MalformedURLException when an exception is returned in an
     *         attempt to construct a java.net.URL object with the supplied
     *         URL string.
     * @throws Exception upon error.
     */
    public HttpClientAPI(String urlString)
        throws MalformedURLException, Exception {

        // QAI 95669: indirect consequence of QAI 93942
        if (urlString == null || urlString.length() < 0) {
            String msg = HTTP_URL_WARNING_MESSAGE2;
            throw new Exception(msg);
        }
        try {
            httpURL = new URL(urlString);
        } catch (MalformedURLException mfue) {
            String msg = "MalformedURLException for url (" + urlString + "): ";
            throw new MalformedURLException(msg + mfue.getMessage());
        }
        this.urlString = new String(urlString);

        // Initialize Header cache and Content-Type values cache
        headerCache = new HashMap();
        contentTypeValsCache = new Vector();
        encoding = "ASCII";

        if (isJDK13only()) {
            if (this.logger.isDebugEnabled() == true) {
                this.logger.debug("Running in 1.3 JVM");
            }
            httpJDKPlatform = (HttpClientAPIJDKPlatform) (Class.forName(
                    "com." + "stc." + "connector." + "httpadapter." + "http."
                    + "jdk13." + "HttpClientAPIJDK13")).newInstance();
        } else {
            if (this.logger.isDebugEnabled() == true) {
                this.logger.debug("NOT running in 1.3 JVM");
            }
            httpJDKPlatform = (HttpClientAPIJDKPlatform) (Class.forName(
                    "com." + "stc." + "connector." + "httpadapter." + "http."
                    + "jdk14." + "HttpClientAPIJDK14")).newInstance();
        }
    }

    /**
     * Sets the SSLContextJDKPlatform for use with HTTPS (SSL)
     *
     * @param sslContext The SSLContextJDKPlatform to use or null to use the
     *        default SSLContextJDKPlatform.  The default
     *        SSLContextJDKPlatform is used if no SSLContextJDKPlatform was
     *        set in the first place.
     */
    public void setSSLContext(SSLContextJDKPlatform sslContext) {
        if (this.logger.isDebugEnabled() == true) {
            this.logger.debug("SSLContext set");
        }
        this.sslContext = sslContext;
    }

    /**
     * Gets the SSLContextJDKPlatform for use with HTTPS (SSL)
     *
     * @return The SSLContextJDKPlatform in use for SSL or null if the default
     *         SSLContextJDKPlatform is used.
     */
    public SSLContextJDKPlatform getSSLContext() {
        return sslContext;
    }

    /**
     * Sets the URL.
     *
     * @param urlString Samples -  http://www.somehost.com
     *        http://www.somehost.com/cgi-bin/query
     *        http://www.somehost.com/cgi-bin/query?name1=value1&name2=value2
     *        The name value pairs that are supplied in the URL must be
     *        urlencoded. The QueryString class is preferred when doing a Post
     *        or Get with mutlitple name value pairs rather than hardcoded in
     *        the URL as in the third example.
     *
     * @throws MalformedURLException when an exception is returned in an
     *         attempt to construct a java.net.URL object with the supplied
     *         URL string.
     */
    public void setURL(String urlString)
        throws MalformedURLException {
        // Remember the old URL
        String oldURL = this.urlString;

        // QAI 95669: indirect consequence of QAI 93942
        if (urlString == null || urlString.length() < 0) {
            String msg = HTTP_URL_WARNING_MESSAGE1 + ".";
            throw new MalformedURLException(msg);
        }
        try {
            // Set the new URL.
            httpURL = new URL(urlString);
        } catch (MalformedURLException mfue) {
            String msg = "MalformedURLException for url (" + urlString + "): ";
            throw new MalformedURLException(msg + mfue.getMessage());
        }
        this.urlString = new String(urlString);

        // Reset if url string is different from last request
        /*
        if ((urlString != null) && (oldURL != null)
                && (!oldURL.equals(urlString))) {
            // Initialize Header cache and Content-Type values cache
            //headerCache = new HashMap ();
            //contentTypeValsCache = new Vector();
            //query           = null;
            //textData        = null;
            //binaryData      = null;
            //outgoingType            = new String(urlOutgoing);
        }
        */
        if (this.logger.isDebugEnabled() == true) {
            this.logger.debug("URL set as " + urlString);
        }
    }

    /**
     * Gets the URL.
     *
     * @return The URL String previously set by setURL.
     *
     * @see #setURL
     */
    public String getURL() {
        return urlString;
    }

    /**
     * Sets the HTTP Authenticator object for use with web sites that require
     * username and password authentication.
     *
     * @param httpAuthenticator The HTTP Authenticator object.
     *
     * @see httpAuthenticator
     */
    public void setHttpAuthenticator(HttpAuthenticator httpAuthenticator) {
        if (httpAuthenticator != null) {
            // Store HttpAuthenticator away.
            this.httpAuthenticator = httpAuthenticator;

            // Register the HttpAuthenticator
            this.httpAuthenticator.register();
            if (this.logger.isDebugEnabled() == true) {
                this.logger.debug("HttpAuthenticator registered for http authentication.");
            }
        }
    }

    /**
     * Gets the current HTTP Authenticator object that was previously set with
     * setHttpAuthenticator method.
     *
     * @return The HTTP Authenticator object that was previously set or null if
     *         not previously set.
     *
     * @see #setHttpProxyHost
     */
    public HttpAuthenticator getHttpAuthenticator() {
        return httpAuthenticator;
    }

    /**
     * Sets the HTTP proxy host.
     *
     * @param httpProxyHost The HTTP proxy host to use.
     *
     * @throws Exception if unable to set HTTP proxy host.
     */
    public void setHttpProxyHost(String httpProxyHost)
        throws Exception {
        try {
            // If user sets proxy, this implies the user wants to use
            // proxy so set use http proxy to true.
            System.setProperty(
                httpProxySetProperty,
                "true");
            System.setProperty(
                httpProxyHostProperty,
                httpProxyHost);
            if (this.logger.isDebugEnabled() == true) {
                this.logger.debug("HTTP Proxy Host set as " + getHttpProxyHost());
            }
        } catch (Exception ex) {
            throw new Exception("Failed to set HTTP Proxy Host; "
                + "Exception : " + ex.toString());
        }
    }

    /**
     * Gets the current HTTP proxy host that was previously set.
     *
     * @return The HTTP proxy host currently being used or null if  HTTP proxy
     *         host was not set.
     *
     * @see #setHttpProxyHost
     */
    public String getHttpProxyHost() {
        return System.getProperty(httpProxyHostProperty);
    }

    /**
     * Sets the HTTP proxy port.
     *
     * @param httpProxyPort The HTTP proxy port to use on the HTTP proxy  host
     *        set with setHttpProxyHost.
     *
     * @throws Exception if unable to set HTTP Proxy Port.
     *
     * @see #setHttpProxyHost
     */
    public void setHttpProxyPort(int httpProxyPort)
        throws Exception {
        try {
            System.setProperty(
                httpProxyPortProperty,
                Integer.toString(httpProxyPort));
            if (this.logger.isDebugEnabled() == true) {
                this.logger.debug("HTTP Proxy Port set as " + getHttpProxyPort());
            }
        } catch (Exception ex) {
            throw new Exception("Failed to set HTTP Proxy Port; "
                + "Exception : " + ex.toString());
        }
    }

    /**
     * Gets the current HTTP proxy port that was previously set.
     *
     * @return The HTTP proxy port currently being used or -1 if  HTTP proxy
     *         host was not set.
     *
     * @see #setHttpProxyPort
     */
    public int getHttpProxyPort() {
        String proxyPort = System.getProperty(httpProxyPortProperty);

        if (proxyPort == null) {
            return -1;
        } else {
            return Integer.parseInt(proxyPort);
        }
    }

    /**
     * Sets the HTTPs proxy host.
     *
     * @param httpsProxyHost The HTTPs proxy host to use.
     *
     * @throws Exception if unable to set HTTPs Proxy Host
     */
    public void setHttpsProxyHost(String httpsProxyHost)
        throws Exception {
        try {
            // If user sets proxy, this implies the user wants to use
            // proxy so set use https proxy to true.
            System.setProperty(
                httpsProxySetProperty,
                "true");
            System.setProperty(
                httpsProxyHostProperty,
                httpsProxyHost);
            if (this.logger.isDebugEnabled() == true) {
                this.logger.debug("HTTPS Proxy Host set as " + getHttpsProxyHost());
            }
        } catch (Exception ex) {
            throw new Exception("Failed to set HTTPs Proxy Host; Exception : "
                + ex.toString());
        }
    }

    /**
     * Gets the current HTTPs proxy host that was previously set.
     *
     * @return The HTTPs proxy host currently being used or null if  HTTPs
     *         proxy host was not set.
     *
     * @see #setHttpsProxyHost
     */
    public String getHttpsProxyHost() {
        return System.getProperty(httpsProxyHostProperty);
    }

    /**
     * Sets the HTTPs proxy port.
     *
     * @param httpsProxyPort The HTTPs proxy port to use on the HTTPs proxy
     *        host set with setHttpsProxyHost.
     *
     * @throws Exception if unable to set HTTPs Proxy Port.
     *
     * @see #setHttpsProxyHost
     */
    public void setHttpsProxyPort(int httpsProxyPort)
        throws Exception {
        try {
            System.setProperty(
                httpsProxyPortProperty,
                Integer.toString(httpsProxyPort));
            if (this.logger.isDebugEnabled() == true) {
                this.logger.debug("HTTPS Proxy Port set as " + getHttpsProxyPort());
            }
        } catch (Exception ex) {
            throw new Exception("Failed to set HTTPs Proxy Port; Exception : "
                + ex.toString());
        }
    }

    /**
     * Gets the current HTTPs proxy port that was previously set.
     *
     * @return The HTTPs proxy port currently being used or -1 if  HTTPs proxy
     *         host was not set.
     *
     * @see #setHttpsProxyPort
     */
    public int getHttpsProxyPort() {
        String proxyPort = System.getProperty(httpsProxyPortProperty);

        if (proxyPort == null) {
            return -1;
        } else {
            return Integer.parseInt(proxyPort);
        }
    }

    /**
     * Set the proxy username for proxy authentication.
     *
     * @param userName The username for proxy authentication.
     */
    public void setProxyUserName(String userName) {
        this.proxyUserName = userName;
        if (this.logger.isDebugEnabled() == true) {
            this.logger.debug("Proxy username set as " + getProxyUserName());
        }
    }

    /**
     * Get the proxy username previously set.
     *
     * @return The username previously set for for proxy authentication or null
     *         if not previously set.
     */
    public String getProxyUserName() {
        return proxyUserName;
    }

    /**
     * Set the proxy password for proxy authentication.
     *
     * @param passWord The password for proxy authentication.
     */
    public void setProxyPassWord(String passWord) {
        this.proxyPassWord = passWord;
        if (this.logger.isDebugEnabled() == true) {
            this.logger.debug("Proxy password set");
        }
    }

    /**
     * Get the proxy password previously set.
     *
     * @return The password previously set for for proxy authentication or null
     *         if not previously set.
     */
    public String getProxyPassWord() {
        return proxyPassWord;
    }

    /**
     * Set the username for http authentication.
     *
     * @param userName The username for http authentication.
     */
    public void setHttpUserName(String userName) {
        this.httpUserName = userName;
        if (this.logger.isDebugEnabled() == true) {
            this.logger.debug("HTTP username set as " + getHttpUserName());
        }
    }

    /**
     * Get the http username previously set.
     *
     * @return The username previously set for for http authentication or null
     *         if not previously set.
     */
    public String getHttpUserName() {
        return httpUserName;
    }

    /**
     * Set the http password for http authentication.
     *
     * @param passWord The password for http authentication.
     */
    public void setHttpPassWord(String passWord) {
        this.httpPassWord = passWord;
        if (this.logger.isDebugEnabled() == true) {
            this.logger.debug("Http password set");
        }
    }

    /**
     * Get the http password previously set.
     *
     * @return The password previously set for http authentication or null
     *         if not previously set.
     */
    public String getHttpPassWord() {
        return httpPassWord;
    }

    /**
     * Enable or disable cookies.
     *
     * @param allowCookie Set to true to allow cookies; otherwise set to false
     *        to disable cookies.
     */
    public void setCookie(boolean allowCookie) {
        this.allowCookie = allowCookie;
        if (this.logger.isDebugEnabled() == true) {
            this.logger.debug("Allow Cookie set as " + allowCookie);
        }
    }

    /**
     * Sets the raw text data to be posted.
     *
     * @param textData The text data to be posted. For example, xml message.
     */
    public void setTextData(String textData) {
        this.textData = textData;
        this.outgoingType = new String(textOutgoing);
        if (this.logger.isDebugEnabled() == true) {
            this.logger.debug("Text data set for POST [" + textData + "]");
        }
    }

    /**
     * Gets the text data to be posted that was previously set by setTextData.
     *
     * @return The text data to be posted.  Returns null if text data was not
     *         set.
     *
     * @see #setTextData
     */
    public String getTextData() {
        return textData;
    }

    /**
     * Sets the raw binary data to be posted.
     *
     * @param binaryData The binary data to be posted.
     */
    public void setBinaryData(byte[] binaryData) {
        if (binaryData == null) {
            this.binaryData = null;
        } else {
            int arraySize = binaryData.length;
            this.binaryData = new byte[arraySize];
            System.arraycopy(
                binaryData,
                0,
                this.binaryData,
                0,
                arraySize);
            this.outgoingType = new String(binaryOutgoing);
            if (this.logger.isDebugEnabled() == true) {
                this.logger.debug("Binary data set for POST");
            }
        }
    }

    /**
     * Gets the binary data to be posted that was previously set by
     * setBinaryData.
     *
     * @return The text data to be posted.  Returns null if binary data was not
     *         previously set.
     *
     * @see #setBinaryData
     */
    public byte[] getBinaryData() {
        byte[] temp = null;

        if (binaryData != null) {
            int arraySize = binaryData.length;

            if (arraySize > 0) {
                temp = new byte[arraySize];
                System.arraycopy(
                    binaryData,
                    0,
                    temp,
                    0,
                    arraySize);
                return temp;
            }
        }

        return temp;
    }

    /**
     * Sets the QueryString for a query
     *
     * @param query The QueryString which contains the url-encoded name value
     *        pairs.
     *
     * @see QueryString
     */
    public void setQueryString(QueryString query) {
        if (query == null) {
            this.query = null;
        } else {
            this.query = (QueryString) query.clone();
            this.outgoingType = new String(formDataOutgoing);
            if (this.logger.isDebugEnabled() == true) {
                this.logger.debug("URL encoded data (QueryString) set as [" + query.toString() + "]");
            }
        }
    }

    /**
     * Gets the query data that was previously set by setQueryString
     *
     * @return The text data to be posted.  Returns null if form data was not
     *         previously set.
     *
     * @see #setQueryString
     */
    public QueryString getQueryString() {
        return query;
    }

    /**
     * POSTs the data previously set by one of the following methods :
     * setTextData setBinaryData setQueryString  The Content-Type can be
     * changed with setContentType() before posting the data.
     *
     * @return an HTTPResponse structure containing the response
     *
     * @throws Exception when an exception is returned from attempting to open
     *         an input stream from the connection.
     *
     * @see #setTextData
     * @see #setBinaryData
     * @see #setQueryString
     * @see HttpResult
     */
    public HttpResult post()
        throws Exception {
        posting = true;

        // QAI 95669: indirect consequence of QAI 93942
        if (urlString == null || urlString.length() < 0) {
            String msg = HTTP_URL_WARNING_MESSAGE1 + ".";
            throw new Exception(msg);
        }
        try {
            // Open new HTTP connection for post
            httpURL = new URL(urlString);
        } catch (MalformedURLException mfue) {
            String msg = "MalformedURLException for url (" + urlString + "): ";
            throw new MalformedURLException(msg + mfue.getMessage());
        }

        if (this.logger.isDebugEnabled() == true) {
            this.logger.debug("Setting URL for POST as " + urlString);
        }
        httpUC = null;

        try {
            httpUC = (HttpURLConnection) httpURL.openConnection();
            if (this.logger.isDebugEnabled() == true) {
                this.logger.debug("URL openConnection succeeded");
            }
           
            /** The default is true for setFollowRedirects();
             **/
            // httpUC.setFollowRedirects(true);
            /**RC set to false so in one session can not auto redirect for us
             * to catch cookie. also the setcookie or getcookie will try to connect to url
             * , so we place after the setheaders call, otherwise the http client will have ilegicalstate
             * exception.**/        
            if (allowCookie) {
                // Use JCookie cookie manager.
                /**setCookies must be called before connection is established
                 ***/
                try {
                    CookieManager.getInstance().setCookies(httpUC);                 
                    if (this.logger.isDebugEnabled() == true) {
                        this.logger.debug("Cookies set for connection: "+ httpUC.getURL().toString());
                    }
                } catch (Exception ex) {
                    this.logger.error("post() processing cookie exception",ex);
                }
               
              
            }
            httpUC.setInstanceFollowRedirects(false);
            
            if (noHostVerification) {
                if (httpJDKPlatform.setNoHostnameVerifier(httpUC)) {
                    if (this.logger.isDebugEnabled() == true) {
                        this.logger.debug("https connection; successfully set " +
                                          "no hostname verfier");
                    }
                } else {
                    if (this.logger.isDebugEnabled() == true) {
                        this.logger.debug("Non-https connection; did not set " +
                                          "no hostname verfier");
                    }
                }
            } else {
                if (this.logger.isDebugEnabled() == true) {
                    this.logger.debug("Hostname verification enabled for https connections.");
                }
            }

            if (!httpJDKPlatform.setTunnelSocketFactory(
                        httpUC,
                        sslContext,
                        getHttpProxyHost(),
                        getHttpProxyPort(),
                        proxyUserName,
                        proxyPassWord)) {
                // Set up SSLSocketFactory if non-default SSLContextJDKPlatform is used
                if (sslContext != null) {
                    if (httpJDKPlatform.setSocketFactory(httpUC, sslContext)) {
                        if (this.logger.isDebugEnabled() == true) {
                            this.logger.debug("SSL tunnel socket factory set the old way");
                        }
                    }
                }
            } else {
                if (this.logger.isDebugEnabled() == true) {
                    this.logger.debug("SSL tunnel socket factory set; proxy host [" +
                                      getHttpProxyHost() + "] proxy port [" + getHttpProxyPort() + "] " +
                                      "proxy username [" + proxyUserName + "]");
                }
            }
            /** RC for cookie during http redirect do not do it here
            httpUC.setFollowRedirects(true);
            httpUC.setInstanceFollowRedirects(true);
            
            if (this.logger.isDebugEnabled() == true) {
                this.logger.debug("Follow redirects set to true");
            }
            **/
        } catch (Exception ex) {
            throw new Exception("Connection can not be opened. "
                + "Got an exception while trying to "
                + "open the URL connection. Exception : " + ex.toString());
        }

        if (httpUC == null) {
            throw new Exception("Connection can not be opened. "
                + "Got a null connection.");
        }

        HttpResult result = null;

        // If user sets all data to null, then assume url-encoded post
        // This is needed to support the case where the user does a post or get
        // with some data and then does a post or get with no data.
        if ((this.textData == null) && (this.query == null)
                && (this.binaryData == null)) {
            try {
               // Set all headers related to this request.
               setHeaders();
            } catch (Exception e) {
               throw new Exception(e.getMessage());
            }
            // Normalize to get
            result = doGet();
        } else { // Do normal, check the type of data being posted
            if (outgoingType.equals(textOutgoing)) {
                // Default to text/plain for Content-Type if no Content-Type was
                // previously given.
                if (this.contentTypeValsCache.size() == 0) {
                    contentTypeValsCache.add("text/plain");
                }

                // Add the Content-Length header
                headerCache.put(
                    CONTENT_LENGTH,
                    (new Integer(textData.length())).toString());

                 try {
                     // Set all headers related to this request.
                     setHeaders();
                 } catch (Exception e) {
                     throw new Exception(e.getMessage());
                 }
                // Post text data
                result = doPostText(
                       // httpUC,
                        textData);
            } else if (outgoingType.equals(formDataOutgoing)) {
                try {
                  // Set all headers related to this request.
                  setHeaders();
                } catch (Exception e) {
                  throw new Exception(e.getMessage());
                }
                // Post Query data (form data)
                result = doPostText(
                       //  httpUC,
                        query.toString());
            } else if (outgoingType.equals(binaryOutgoing)) {
                // Default to application/octet-stream for Content-Type
                // if no Content-Type was previously given.
                if (this.contentTypeValsCache.size() == 0) {
                    contentTypeValsCache.add("application/octet-stream");
                }

                // Add the Content-Length header
                headerCache.put(
                    CONTENT_LENGTH,
                    (new Integer(binaryData.length)).toString());
                try {
                   // Set all headers related to this request.
                   setHeaders();
                } catch (Exception e) {
                  throw new Exception(e.getMessage());
                }
                // Post binary data
                result = doPostBinary(
                        // httpUC,
                        binaryData);
            } else if (outgoingType.equals(urlOutgoing)) {
                try {
                   // Set all headers related to this request.
                   setHeaders();
                } catch (Exception e) {
                  throw new Exception(e.getMessage());
                }
                // Normalize to get
                result = doGet();
            }
        }

        // If we got redirected from http to https, restore the original
        // URL the user provided
        if (savedURLOnRedirect != null) {
            urlString = savedURLOnRedirect;
            savedURLOnRedirect = null;
        }

        return result;
    }

    /**
     * GETs the data previously set by the following methods : setURL
     * setQueryString
     *
     * @return an HttpResponse structure containing the response
     *
     * @throws Exception when an exception is returned from attempting to open
     *         an input stream from the connection.
     *
     * @see #setURL
     * @see #setQueryString
     * @see HttpResult
     */
    public HttpResult get()
        throws Exception {
        posting = false;

        httpUC = null;

        // QAI 95669: indirect consequence of QAI 93942
        if (urlString == null || urlString.length() < 0) {
            String msg = HTTP_URL_WARNING_MESSAGE2;
            throw new Exception(msg);
        }
        try {
            URL testURL = new URL(urlString);
        } catch (MalformedURLException mfue) {
            String msg = "MalformedURLException for url (" + urlString + "): ";
            throw new MalformedURLException(msg + mfue.getMessage());
        }

        // If user sets all data to null, then assume url-encoded get
        // This is needed to support the case where the user does a post or get
        // with some data and then does a post or get with no data.
        if ((this.textData == null) && (this.query == null)
                && (this.binaryData == null)) {
            // Create new HTTP URL with url get
            httpURL = new URL(urlString);
            if (this.logger.isDebugEnabled() == true) {
                this.logger.debug("Setting URL for GET as " + urlString);
            }
        } else { // Do normal, check the type of data being posted
            if (outgoingType.equals(urlOutgoing)) {
                // Create new HTTP URL with url get
                httpURL = new URL(urlString);
                if (this.logger.isDebugEnabled() == true) {
                    this.logger.debug("Setting URL for GET as " + urlString);
                }
            } else if (outgoingType.equals(formDataOutgoing)) {
                // Create new HTTP URL for url encoded form get
                httpURL = new URL(urlString + "?" + query.toString());
                if (this.logger.isDebugEnabled() == true) {
                    this.logger.debug("Setting URL for GET as " + urlString +
                                      "?" + query.toString());
                }
            }
        }

        try {
            // Open the HTTP connection
            httpUC = (HttpURLConnection) httpURL.openConnection();          
            if (this.logger.isDebugEnabled() == true) {
                this.logger.debug("URL openConnection succeeded");
            }
           /** The default is true for setFollowRedirects();
             **/
            // httpUC.setFollowRedirects(true);
            /**RC set to false so in one session can not auto redirect for us
             * to catch cookie. also the setcookie or getcookie will try to connect to url
             * , so we place after the setheaders call, otherwise the http client will have ilegicalstate
             * exception.**/        
            if (allowCookie) {
                // Use JCookie cookie manager.
                /**setCookies must be called before connection is established
                 ***/
                try {
                    CookieManager.getInstance().setCookies(httpUC);
                    if (this.logger.isDebugEnabled() == true) {
                        this.logger.debug("Cookies set for connection: "+ httpUC.getURL().toString());
                    }
                } catch (Exception ex) {
                    this.logger.error("post() processing cookie exception",ex);
                }
               
              
            }
            httpUC.setInstanceFollowRedirects(false);
            if (this.logger.isDebugEnabled() == true) {
                //this.logger.debug("Follow redirects set to true");
                this.logger.debug("but InstanceFollowRedirects  set to false");
            }
            
            if (noHostVerification) {
                if (httpJDKPlatform.setNoHostnameVerifier(httpUC)) {
                    if (this.logger.isDebugEnabled() == true) {
                        this.logger.debug("https connection; successfully set " +
                                          "no hostname verfier");
                    }
                } else {
                    if (this.logger.isDebugEnabled() == true) {
                        this.logger.debug("Non-https connection; did not set " +
                                          "no hostname verfier");
                    }
                }
            } else {
                if (this.logger.isDebugEnabled() == true) {
                    this.logger.debug("Hostname verification enabled for https connections.");
                }
            }

            if (!httpJDKPlatform.setTunnelSocketFactory(
                        httpUC,
                        sslContext,
                        getHttpProxyHost(),
                        getHttpProxyPort(),
                        proxyUserName,
                        proxyPassWord)) {
                // Set up SSLSocketFactory if non-default SSLContext is used
                if (sslContext != null) {
                    if (httpJDKPlatform.setSocketFactory(httpUC, sslContext)) {
                        this.logger.debug("SSL tunnel socket factory set the old way");
                    }
                }
            } else {
                if (this.logger.isDebugEnabled() == true) {
                    this.logger.debug("SSL tunnel socket factory set; proxy host [" +
                                      getHttpProxyHost() + "] proxy port [" + getHttpProxyPort() + "] " +
                                      "proxy username [" + proxyUserName + "]");
                }
            }

           
        } catch (Exception ex) {
            throw new Exception("Connection can not be opened. "
                + "Got an exception while trying to open the URL connection. "
                + "Exception : "
                + ex.toString());
        }

        if (httpUC == null) {
            throw new Exception(
                "Connection can not be opened. Got a null connection.");
        }
        try {
           // Set all headers related to this request.
           setHeaders();
        } catch (Exception e) {
            throw new Exception(e.getMessage());
        }
        // Get the data.
        HttpResult result = doGet();

        // If we got redirected from http to https,
        // restore the original URL the user provided
        if (savedURLOnRedirect != null) {
            urlString = savedURLOnRedirect;
            savedURLOnRedirect = null;
        }

        return result;
    }

    /**
     * Adds the specified header such that the next request sent will contain
     * the specified header information. If the specified header was already
     * previously added, then nothing is done.
     *
     * @param header The header to be added.
     *
     * @see HttpHeader
     */
    public void addHeader(HttpHeader header) {
        headerCache.put(
            new String(header.getName()),
            new String(header.getValue()));
    }

    /**
     * Adds the specified header such that the next request sent will contain
     * the specified header information. If the specified header was already
     * previously added, then nothing is done.
     *
     * @param name The name portion of the header to be added.
     * @param value The value portion of the header to be added.
     */
    public void addHeader(String name, String value) {
        headerCache.put(
            new String(name),
            new String(value));
    }

    /**
     * If the specified header was previously set, this method removes the
     * header such that the next request sent will not contain the specified
     * header information.  Otherwise, nothing is done.
     *
     * @param header The header to remove from the list of headers.
     */
    public void clearHeader(HttpHeader header) {
        headerCache.remove(header.getName());
    }

    /**
     * If the specified header was previously set, this method removes the
     * header such that the next request sent will not contain the specified
     * header information.  Otherwise, nothing is done.
     *
     * @param name The name of the header used to locate the HTTP header for
     *        removal from the list of headers.
     */
    public void clearHeader(String name) {
        headerCache.remove(name);
    }

    /**
     * Removes all previously added headers such that the next request sent
     * will not contain any of the previously added header information.  If no
     * headers were previously added, nothing will be done.
     */
    public void clearHeaders() {
        headerCache.clear();
    }

    /**
     * Get the count of headers previously added.
     *
     * @return The count of HTTP headers previously added or 0 if no headers.
     */
    public int headerCount() {
        if (headerCache != null) {
            return headerCache.size();
        } else {
            return 0;
        }
    }

    /**
     * Adds a Content-Type value such that the next request sent will contain
     * the specified value in the Content-Type header.
     *
     * @param contentType The String specifying the type of message content of
     *        the request (i.e., text/html).
     */
    public void addContentType(String contentType) {
        contentTypeValsCache.add(contentType);
    }

    /**
     * If the specified Content-Type was previously added, this method removes
     * the Content-Type value such that the next request sent  will not
     * contain the specified Content-Type value.   Otherwise, nothing is done.
     *
     * @param contentType The Content Type to remove.
     */
    public void clearContentType(String contentType) {
        contentTypeValsCache.remove(contentType);
    }

    /**
     * Get the count of content types previously added.
     *
     * @return The count of content types previously added or 0 if no content
     *         types.
     */
    public int contentTypeCount() {
        if (contentTypeValsCache != null) {
            return contentTypeValsCache.size();
        } else {
            return 0;
        }
    }

    /**
     * Removes all previously added Content-Type values such that  the next
     * request sent will not contain any of the previously added Content-Type
     * information.  If no headers were previously added, nothing will be
     * done.
     */
    public void clearContentTypes() {
        contentTypeValsCache.removeAllElements();
    }

    /**
     * Clears all headers and request data from memory.
     */
    public void reset() {
        query = null;
        textData = null;
        binaryData = null;
        outgoingType = urlOutgoing;
        urlString = null;
        clearContentTypes();
        clearHeaders();
    }

    /**
     * Set the encoding.
     *
     * @param encoding String for encoding (i.e., ASCII).
     */
    public void setEncoding(String encoding) {
        this.encoding = encoding;
        URLEncoder2.setEncoding(encoding);
    }

    /**
     * Get the encoding.
     *
     * @return String for encoding (i.e., ASCII) that was previously set.
     */
    public String getEncoding() {
        return encoding;
    }

    /**
     * Set the flag for hostname verification on https connections.
     *
     * @param noHostVerification    Set to true if no hostname verification
     *                              is desired for https connections; false otherwise.
     */
    public void setNoHostnameVerifier(boolean noHostVerification) {
        this.noHostVerification = noHostVerification;
    }

    /**
     * Get the the flag for hostname verification on https connections.
     *
     * @return True if hostname verification is turned off; false otherwise.
     */
    public boolean getNoHostnameVerifier() {
        return noHostVerification;
    }

    //============================//
    // private, private methods //
    //============================//
    // Add Content-Type header (may have multiple values)
    // The values, of a multi-value Content-Type header, are separated
    // by a comma.
    private void setContentTypeHeader() {
        int contentCacheSize = contentTypeValsCache.size();

        if (contentCacheSize > 0) {
            StringBuffer contentVal = new StringBuffer();

            for (int i = 0; i < contentTypeValsCache.size(); i++) {
                if (i == 0) {
                    contentVal.append((String) contentTypeValsCache.get(i));
                } else {
                    contentVal.append(
                        "," + (String) contentTypeValsCache.get(i));
                }
            }

            httpUC.setRequestProperty(
                CONTENT_TYPE,
                contentVal.toString());
            if (this.logger.isDebugEnabled() == true) {
                this.logger.debug("will Set " + CONTENT_TYPE
                    + " header to " + contentVal.toString());       
            }
        }
    }

    // Add any other user specified headers
    private void setUserSpecifiedHeaders() {
        HttpHeader traverseHeader;
        if (!this.beingRedirected) {
            /**
             *Repost on http response redirect should not set header secand times
             */
            if (headerCache.size() > 0) {
                Iterator keyIter = headerCache.keySet().iterator();
                
                while (keyIter.hasNext()) {
                    String name = (String) keyIter.next();
                    String value = (String) headerCache.get(name);
                    httpUC.setRequestProperty(
                       name,
                       value);
                    if (this.logger.isDebugEnabled() == true) {
                        this.logger.debug("Header will be set; name [" + name
                        + "] value [" + value +"]");                       
                    }
                }
            }
        }
    }

    // Add http user authentication header
    private void setAuthorizationHeader () throws Exception {
      
        // We should be able to automatically detect whether basic or digest
        // is to be used by finding out whether digest or basic header was received.

        if (this.httpUserName != null && this.httpUserName.length() > 0) {

            if (this.authSchemeToAttempt == BASIC_AUTH) {

               // Current Strongest auth scheme level needs to be higher than
               // what is detected to move on
               if (this.authScheme >= BASIC_AUTH) {
                  
                  // I guess it is possible that the password is empty so we are not
                  // going to validate user name
                  
                  String userPassword = this.httpUserName + ":" + this.httpPassWord;      
                  String encoding = null;
                  try {
                      encoding = new BASE64Encoder().encode(userPassword.getBytes());
                  } catch (Exception ex) {
                      this.logger.error("Error in encoding password with BASE64Encoder().encode",ex);
                  }
                  httpUC.setRequestProperty ("Authorization",
                                                     "Basic " + encoding);
                  if (this.logger.isDebugEnabled() == true) {
                      this.logger.debug("Authorization header set to [Basic "
                                          + encoding + "]");
                  }
               } // end "if (this.authScheme >= BASIC_AUTH) {"

            } else if (this.authSchemeToAttempt == DIGEST_AUTH) {
            
               // Current Strongest auth scheme level needs to be higher than
               // what is detected to move on
               if (this.authScheme >= DIGEST_AUTH) {
   
                  // QAI 90061 -- Digest Authentication
                  //
                  // Sample challenge from Tomcat server:
                  //
                  // Digest realm="private area protected by Digest", nonce="HK78go4CBAA=a13d2d7e7e95f8aeaacd994cff1c17501fedc449", algorithm=MD5, qop="auth"
                  //
                  // Unfortunately I was told that HttpAuthenticator may not work with SSL, proxy, etc so the
                  // following simple mechanism would not work for all scenario:
                  //
                  //    HttpAuthenticator aHttpAuthenticator = 
                  //       new HttpAuthenticator(this.httpUserName, this.httpPassWord);
                  //    setHttpAuthenticator(aHttpAuthenticator);
                  //
                  if (this.responseCode == UNAUTHORIZED_RESPONSE_CODE) {
   
                     this.responseCode = SUCCESS_RESPONSE_CODE; // Important to avoid recursive trap
   
                     String credentials = "";
                     try {
                        credentials = getDigestAuthorizationCredentials(this.httpUserName,
                                                                        this.httpPassWord);
                     } catch (Exception e) {
                        throw new Exception(e.getMessage());
                     }
   
                     httpUC.setRequestProperty ("Authorization", credentials);
   
                     if (this.logger.isDebugEnabled() == true) {
                         this.logger.debug("Authorization header set to [Digest "
                                             + credentials + "]");
                     }
                  } else {
                     // Important to reset the digest challenge retry if 401 is no longer received
                     // to avoid recursive trap
                     //
                     this.challenge_retry = 0; 

                  } // end "// end "if (this.responseCode == UNAUTHORIZED_RESPONSE_CODE) ""

               } // end "if (this.authScheme >= DIGEST_AUTH) {"

            } else {
               // TODO: actually it shouldn't even be AUTO_AUTH at this point
               if (this.authSchemeToAttempt != AUTO_AUTH) {
                  String msg = "Unknown authentication scheme code [" + this.authSchemeToAttempt + "]";
                  this.logger.error(msg);
                  // TODO: might want to make stop on error (with exception) configurable;
                  // some user might not care but some might do
                  //throw new Exception(msg); 
               }
            } // end "if (this.authSchemeToAttempt == BASIC_AUTH) {"

        } else {
            // QAI 95669 -- it is OK without the http user name and password for 
            // basic/digest/none-of-these-2 http operation.
            /*
            if ( (this.authSchemeToAttempt == BASIC_AUTH) ||
                 (this.authSchemeToAttempt == DIGEST_AUTH) ){
               //
               // Missing user name but there was indeed a challenge received 
               // for the underlined URL set
               //
               String msg = "Missing credential user name for basic or digest authentication; ";
                      msg += "please supply one in configuration entry or make use of the setCredentials method in user collaboration";
               this.logger.error(msg);
               // TODO: might want to make stop on error (with exception) configurable;
               // some user might not care but some might do
               //throw new Exception(msg); 
            }
            */
        } // end "if (this.httpUserName != null && this.httpUserName.length() > 0) "
    }

    // Add proxy authentication header - non ssl.  If ssl, then proxy tunnel
    // factory will be used.
    private void setProxyAuthorizationHeader () {
        if (this.proxyUserName != null && this.proxyUserName.length() > 0) {
            String userPassword = this.proxyUserName + ":" + this.proxyPassWord;
            String encoding = null;
            try {
                encoding = new BASE64Encoder().encode(userPassword.getBytes());
            } catch (Exception ex) {
                 this.logger.error("setProxyAuthorizationHeader: Error in encoding password with BASE64Encoder().encode",ex);
            }
            httpUC.setRequestProperty ("Proxy-Authorization",
                                       "Basic " + encoding);
            if (this.logger.isDebugEnabled() == true) {
                this.logger.debug("Proxy-Authorization header will be set to [Basic "
                                + encoding + "]");          
            }
        }
    }

    // Add all header information
    private void setHeaders() throws Exception {
        // Set the Content-Type header
        setContentTypeHeader();

        // Set other headers specified
        setUserSpecifiedHeaders();

        // Set authorization header
        setAuthorizationHeader ();

        // Set proxy-authorization header
        setProxyAuthorizationHeader ();
    }

    // Do the post request for text data
    private HttpResult doPostText( String postData)
        throws IOException {
        try {             /**RC don't send POST data to inaproprate redirects -- see RFC2616 section 10.3
              *we only do a connect
            **/
            if (!this.beingRedirected) {
                // Configure connection for output/input
                httpUC.setDoOutput(true);
                httpUC.setDoInput(true);
                httpUC.setRequestMethod(this.requestPost);
                httpUC.connect();
                // Write the data
                BufferedWriter out = null;
                if (encoding.length() == 0 || encoding.equals("ASCII"))
                    out = new BufferedWriter(new OutputStreamWriter(httpUC.getOutputStream()));
                else
                    out = new BufferedWriter(new OutputStreamWriter(httpUC.getOutputStream(), encoding));
                if (this.logger.isDebugEnabled() == true) {
                    this.logger.debug("Got ouputstream for writing from connection; encoding = " + encoding);
                }
                out.write(postData);
                out.flush();
                out.close();
                if (this.logger.isDebugEnabled() == true) {
                    this.logger.debug("Text data posted successfully");
                    this.logger.debug("Receiving POST response");
                }
            }else {
                httpUC.connect();
            }
            return getResult();
        } catch (IOException ioe) {
            this.logger.error ("Error POSTing text request", ioe);
            throw ioe;
        }
    }

    // Do the post request for text data
    private HttpResult doPostBinary( byte[] postData)
        throws IOException {
        try {
              /**RC don't send POST data to inaproprate redirects -- see RFC2616 section 10.3
              *we only do a connect
            **/
             if (!this.beingRedirected) {
            // Configure connection for output/input
                 httpUC.setDoOutput(true);
                 httpUC.setDoInput(true);
                 httpUC.setRequestMethod(this.requestPost);
                 httpUC.connect();
            // Write the data
                 BufferedOutputStream out = new BufferedOutputStream(
                 httpUC.getOutputStream());
                 if (this.logger.isDebugEnabled() == true) {
                     this.logger.debug("Got ouputstream for writing from connection");
                 }
                 out.write(postData);
                 out.flush();
                 out.close();
                 if (this.logger.isDebugEnabled() == true) {
                     this.logger.debug("Binary (byte[]) data posted successfully");
                     this.logger.debug("Receiving POST response");
                 }
             }else {
                 httpUC.connect();
             }
            return getResult();
        } catch (IOException ioe) {
            this.logger.error ("Error POSTing binary request", ioe);
            throw ioe;
        }
    }

    // Do the get request
    private HttpResult doGet()
        throws IOException {
        try {
            if (this.logger.isDebugEnabled() == true) {
                this.logger.debug("Receiving GET response");
            }
            return getResult();
        } catch (IOException ioe) {
            this.logger.error ("Error while sending GET request", ioe);
            throw ioe;
        }
    }

    // Get a result from the request and return the result as
    // an instance of HttpResult.
    private HttpResult getResult()
        throws IOException {
        Vector headers = new Vector();
        String contentType = null;
        HttpResult result = new HttpResult();
        int contentLength = 0;
        String contentLengthStr = null;

        try {
            
            // Get the headers first
            for (int j = 1; ; j++) {
                String value = httpUC.getHeaderField(j);

                //if (value == null)
                //    break;  // No more headers
                // Retrieve value for the header.
                String name = httpUC.getHeaderFieldKey(j);

                if ( (name == null) && (value == null)) {
                    break;
                }
                if (this.logger.isDebugEnabled() == true) {
                    this.logger.debug("Header received; name [" + name
                                      + "] value [" + value + "]");
                }
                // Set Content-type for reading appropriate data later
                // (text, binary?)
                if (CONTENT_TYPE.equalsIgnoreCase(name)) {
                    contentType = new String(value.toLowerCase());
                }

                // Set Content_length
                if (CONTENT_LENGTH.equalsIgnoreCase(name)) {
                    contentLengthStr = value;
                    contentLength = Integer.parseInt(value);
                }

                // QAI 90061 -- Digest Authentication
                if (WWW_AUTHENTICATE.equalsIgnoreCase(name)) {
                    this.auth_server_challenge = value;

                    // TODO: simulataneous multiple challenges 
                    // (at different levels) are not supported
                    // for simplicity.
                    //
                    if (parse_auth_challenge(this.auth_server_challenge) 
                           >= WWW_AUTHENTICATE_DIGEST_MIN_FIELD_COUNT) {
                        this.authSchemeToAttempt = DIGEST_AUTH;
                    } else {
                        this.authSchemeToAttempt = BASIC_AUTH;
                    } // end "if (parse_auth_challenge(this.auth_server_challenge) "
                }

                if ( (name != null) && (value == null)) {
                    value = new String("");
                }
                else if ( (value != null) && (name == null)) {
                    name = new String("");
                }

                // Store header away
                headers.add(new HttpHeader(
                    name,
                    value));
            }
        } catch (Throwable t) {
            logger.info("Error trying to read response headers; perhaps none were returned by the server. Attempt to read message body.", t);
        }
        try {
            // Set the headers in the http response object that will be returned.        
            result.setHeaders(headers);
        } catch (Exception e) {
            throw new IOException(e.getMessage());
        }
        // Set up SSL Socket Factory if https proxy is set up and we
        // are directed from http to https site.
        // Check to see if we got a redirect to SSL web site
        //URL url = httpC.getURL();
        URL url;
        /**RC since I set instanceredirect to false, this is not going to return redirected url
         *must look at response code to determine if it is redirect
         **/
        //String urlRedirect = url.toString();
        String urlRedirect;
        /**
         *seems like the getResponseCode will causing connect
         */
        int code = httpUC.getResponseCode();
        if (this.logger.isDebugEnabled() == true) {
            this.logger.debug("responce code ["
            + code + "] 3xx is http redirect ");
        };

         /**getCookies must be called after connection is established
          ***/
         if (allowCookie) {
             try {
                 CookieManager.getInstance().getCookies(httpUC);
                 if (this.logger.isDebugEnabled() == true) {
                     this.logger.debug("Cookies get for connection: "+ httpUC.getURL().toString());
                 }
             } catch (Exception ex) {
                 this.logger.error("post() processing cookie exception",ex);
             }
         }
               
        // JDK will put redirected url in HttpURLConnection.URL
        if (code >=300 && code <= 399 ) {

            // QAI 95669: indirect consequence of QAI 93942
            try {
               url = new URL(httpUC.getURL(),httpUC.getHeaderField("location"));
            } catch (MalformedURLException mfue) {
               String msg = "MalformedURLException for url (" + httpUC.getURL() + 
                           ") and header field (" + httpUC.getHeaderField("location") + "): ";
               throw new MalformedURLException(msg + mfue.getMessage());
            }
            urlRedirect = url.toString();
            this.logger.info("Connection redirected from URL  [" + urlString + "] to URL [" +
                             urlRedirect + "]");
            // Check if we are redirected to https url
            String reDirectProtocol = url.getProtocol();
            String origProtocol = httpURL.getProtocol();

            if ( ("http".equalsIgnoreCase(origProtocol)
                    && "https".equalsIgnoreCase(reDirectProtocol)) ||
                   needToSetHeaders()) {
                if (this.logger.isDebugEnabled() == true) {
                    this.logger.debug("Connection redirected to secure URL ["
                                    + urlRedirect + "]");
                }
            }
            // We are in this code because we got redirected from an http to
            // https web site and https proxy is set up... better set up
            // SSL Tunnel Socket factory or else it will hang!
            // First, disconnect previous http proxy connection
            httpUC.disconnect();
            if ( code != 300 && code != 301 && code != 305) {
                this.beingRedirected = true;
                
            }
                // For some reason, disconnect won't disconnect unless we sleep
                try {
                    //Thread.sleep(100);
                    Thread.currentThread().sleep(100);
                } catch (Exception ex) {
                    ;
                }

                // Reset the http URL to redirected url and then start all over
                // But save the original URL
                this.savedURLOnRedirect = this.urlString;
                this.urlString = urlRedirect;

                // Start post or get again with new https url
                try {
                    if (posting) {
                        if (this.logger.isDebugEnabled() == true) {
                            this.logger.debug("Re-POST with URL [" + urlString + "]");
                        }
                        return this.post();
                    } else {
                        if (this.logger.isDebugEnabled() == true) {
                            this.logger.debug("Re-GET with URL [" + urlString + "]");
                        }
                        return this.get();
                    }
                } catch (Exception ex) {
                    throw new IOException(ex.toString());
                }
            }

        // QAI 90061 -- Digest Authentication
        this.responseCode = httpUC.getResponseCode();
        if (httpUC.getResponseCode() == UNAUTHORIZED_RESPONSE_CODE) {

           // TODO: MAX_CHALLENGE_RETRY could be configurable but larger or equal to 1
           if (this.challenge_retry < MAX_CHALLENGE_RETRY) {
              this.challenge_retry++;

              // Start post or get again with new authorization request header
              try {
                if (posting) {
                   if (this.logger.isDebugEnabled() == true) {
                       this.logger.debug("Digest authentication response re-POST with URL [" + urlString + "]: attempt #" + this.challenge_retry);
                   }
                   return this.post();
                } else {
                   if (this.logger.isDebugEnabled() == true) {
                       this.logger.debug("Digest authentication response re-GET with URL [" + urlString + "]: attempt #" + this.challenge_retry);
                   }
                   return this.get();
                }
              } catch (Exception ex) {
                   throw new IOException(ex.toString());
              }
           } else {
               String msg = "Maximum number of " + MAX_CHALLENGE_RETRY + " retries to access URL [" + urlString + "] has been reached; ";
                      msg += "401 unauthorized response codes were received; ";
                      msg += "please make sure that the credentials (user name and password) are correct and up to date";
               this.logger.error(msg);
               // TODO: might want to make stop on error (with exception) configurable;
               // some user might not care but some might do
               //throw new IOException(msg); 

           } // end "if (this.challenge_retry < MAX_CHALLENGE_RETRY) {"
        }
         
        // Now get the data depending on Content_type.
        // Assume text if Content-type is not given.
        if (contentType != null) {
            // Get text data
            if (isTextResult(contentType)) {
                if (this.logger.isDebugEnabled() == true) {
                    this.logger.debug("Got text result");
                }
                StringBuffer response = new StringBuffer();

                BufferedReader in;
                if ((contentLength > 0) || (contentLengthStr == null)) {
                    char[] dataChunk = new char[BUFF_SIZE];

                    try {
                        if ( (encoding.length() == 0) ||
                            encoding.equals("ASCII")) {
                            in = new BufferedReader(
                                new InputStreamReader(httpUC.getInputStream()));
                        }
                        else {
                            in = new BufferedReader(
                                new InputStreamReader(
                                httpUC.getInputStream(),
                                encoding));
                        }
                    } catch (IOException ioex) {
                        if (! getFromErrorStream (result) ) {
                            this.logger.error("IOException caught while trying to get the input stream from the connection.", ioex);
                            throw ioex;
                        } else {
                            return result;
                        }
                    }

                    if (this.logger.isDebugEnabled() == true) {
                        this.logger.debug("Got inputstream for reading from connection; encoding = " + encoding);
                    }

                    int dataRead = in.read(dataChunk);

                    while (dataRead != -1) {
                        response.append(
                            dataChunk,
                            0,
                            dataRead);
                        dataRead = in.read(dataChunk);
                    }
                    if (this.logger.isDebugEnabled() == true) {
                        this.logger.debug("Got response [" + response.toString() + "]");
                    }
                    in.close();
                }

                // Set the data type as text in the http response object that
                // will be returned.
                result.setIsTextResult(true);

                // Set the text data in the http response object that will be
                // returned.
                if (response.length() > 0) {
                    result.setTextResult(response.toString());
                }

                // Set the response code and response message in the http
                // response object that will be returned.
                int rcode = httpUC.getResponseCode();
                this.responseCode = rcode; // QAI 90061 -- Digest Authentication
                String rmsg = httpUC.getResponseMessage();
                result.setResponseCode(rcode);
                result.setResponseMessage(rmsg);
                if (this.logger.isDebugEnabled() == true) {
                    this.logger.debug("Got response code [" + rcode + "]");
                    this.logger.debug("Got response message [" + rmsg + "]");
                }

                // If content returned contains <META HTTP-EQUIV="Refresh" ...>
                // Then set redirected URL and start post or get over
                String metaRedirectURL = checkAndGetMetaRedirectURL(
                        response.toString());

                if (metaRedirectURL != null) {
                    this.logger.info("Got meta data redirect message; redirected URL [" + metaRedirectURL + "]");

                    // But first, save the original URL
                    this.savedURLOnRedirect = this.urlString;
                    this.urlString = metaRedirectURL;

                    // Start post or get again with new https url
                    try {
                        if (posting) {
                            if (this.logger.isDebugEnabled() == true) {
                                this.logger.debug("Re-POST with URL [" + urlString + "]");
                            }
                            return this.post();
                        } else {
                            if (this.logger.isDebugEnabled() == true) {
                                this.logger.debug("Re-GET with URL [" + urlString + "]");
                            }
                            return this.get();
                        }
                    } catch (Exception ex) {
                        throw new IOException(ex.toString());
                    }
                }
            } else { // Everything else is binary

                byte[] data = null;
                if (this.logger.isDebugEnabled() == true) {
                    this.logger.debug("Got binary (byte []) result");
                }
                if (contentLength > 0) {
                    if (this.logger.isDebugEnabled() == true) {
                        this.logger.debug("Content-length of bytes response [" + contentLength + "]");
                    }

                    BufferedInputStream in;
                    try {
                        in = new BufferedInputStream(
                            httpUC.getInputStream());
                    } catch (IOException ioex) {
                        if (! getFromErrorStream (result) ) {
                            this.logger.info("IOException caught while trying to get the input stream from the connection.", ioex);
                            throw ioex;
                        } else {
                            return result;
                        }
                    }

                    if (this.logger.isDebugEnabled() == true) {
                        this.logger.debug("Got inputstream for reading from connection");
                    }
                     /**RC not needed to do here
                      already done in get() post() before the http response redirect
                      * if (allowCookie) {
                      * // Use JCookie cookie manager.
                      * try {
                      * CookieManager.getInstance().getCookies(httpC);
                      * if (this.logger.isDebugEnabled() == true) {
                      * this.logger.debug("Got potential cookies");
                      * }
                      * } catch (Exception ex) {
                      * ex.printStackTrace();
                      * }
                      * }
                      **/

                    data = new byte[contentLength];

                    int bytesRead = 0;
                    int offset = 0;

                    while (offset < contentLength) {
                        // Read until no more data to read
                        bytesRead = in.read(
                                data,
                                offset,
                                data.length - offset);

                        if (bytesRead == -1) {
                            break;
                        }

                        offset += bytesRead;
                    }

                    in.close();
                    if (this.logger.isDebugEnabled() == true) {
                        this.logger.debug("Read all bytes response");
                    }

                    // Make sure we got all of the data as specified in
                    // Content_length.
                    if (offset != contentLength) {
                        throw new IOException("Only read " + offset
                            + " bytes; Expected " + contentLength + " bytes.");
                    }
                } else if (contentLengthStr == null) {  // If no content-length
                    if (this.logger.isDebugEnabled() == true) {
                        this.logger.debug("No Content-length specified for bytes response");
                    }

                    BufferedInputStream in;
                    try {
                        in = new BufferedInputStream(
                            httpUC.getInputStream());
                    } catch (IOException ioex) {
                        if (! getFromErrorStream (result) ) {
                            this.logger.info("IOException caught while trying to get the input stream from the connection.", ioex);
                            throw ioex;
                        } else {
                            return result;
                        }
                    }

                    if (this.logger.isDebugEnabled() == true) {
                        this.logger.debug("Got inputstream for reading from connection");
                    }

                     /**RC not needed to do here
                      already done in get() post() before the http response redirect
                      * if (allowCookie) {
                      * // Use JCookie cookie manager.
                      * try {
                      * CookieManager.getInstance().getCookies(httpC);
                      * if (this.logger.isDebugEnabled() == true) {
                      * this.logger.debug("Got potential cookies");
                      * }
                      * } catch (Exception ex) {
                      * ex.printStackTrace();
                      * }
                      * }
                      **/

                    ByteArrayOutputStream bos = new ByteArrayOutputStream ();
                    byte [] buffer = new byte [BUFF_SIZE];
                    while (true) {
                        int bytesRead = in.read (buffer);
                        if (bytesRead == -1) {
                            break;
                        }
                        bos.write (buffer, 0, bytesRead);
                    }
                    if (bos.size() > 0) {
                        data = bos.toByteArray();
                    }
                    in.close();
                    if (this.logger.isDebugEnabled() == true) {
                        this.logger.debug("Read all bytes response");
                    }
                }

                // Set the data type as binary in the http response object that
                // will be returned.
                result.setIsTextResult(false);

                // Set the binary data in the http response object that will
                // be returned.
                if (data != null) {
                    result.setBinaryResult(data);
                }

                // Set the response code and response message in the http
                // response object that will be returned.
                int rcode = httpUC.getResponseCode();
                this.responseCode = rcode; // QAI 90061 -- Digest Authentication
                String rmsg = httpUC.getResponseMessage();
                result.setResponseCode(rcode);
                result.setResponseMessage(rmsg);
                if (this.logger.isDebugEnabled() == true) {
                    this.logger.debug("Got response code [" + rcode + "]");
                    this.logger.debug("Got response message [" + rmsg + "]");
                }

            }
        } else { // Assumes text content type if Content-type is not set.
            if (this.logger.isDebugEnabled() == true) {
                this.logger.debug("No Content-type; getting text result");
            }
            StringBuffer response = new StringBuffer();

            if ((contentLength > 0) || (contentLengthStr == null)) {
                char[] dataChunk = new char[BUFF_SIZE];
                BufferedReader in;

                try {
                    if ( (encoding.length() == 0) || encoding.equals("ASCII")) {
                        in = new BufferedReader(
                            new InputStreamReader(httpUC.getInputStream()));
                    }
                    else {
                        in = new BufferedReader(
                            new InputStreamReader(
                            httpUC.getInputStream(),
                            encoding));
                    }
                } catch (IOException ioex) {
                    if (! getFromErrorStream (result) ) {
                        this.logger.info("IOException caught while trying to get the input stream from the connection.", ioex);
                        throw ioex;
                    } else {
                        return result;
                    }
                }

                if (this.logger.isDebugEnabled() == true) {
                    this.logger.debug("Got inputstream for reading from connection; encoding = " + encoding);
                }
                /**RC not needed to do here
                      already done in get() post() before the http response redirect
                      * if (allowCookie) {
                      * // Use JCookie cookie manager.
                      * try {
                      * CookieManager.getInstance().getCookies(httpC);
                      * if (this.logger.isDebugEnabled() == true) {
                      * this.logger.debug("Got potential cookies");
                      * }
                      * } catch (Exception ex) {
                      * ex.printStackTrace();
                      * }
                      * }
                      **/

                int dataRead = in.read(dataChunk);

                while (dataRead != -1) {
                    response.append(
                        dataChunk,
                        0,
                        dataRead);
                    dataRead = in.read(dataChunk);
                }
                if (this.logger.isDebugEnabled() == true) {
                    this.logger.debug("Got response [" + response.toString() + "]");
                }
                in.close();
            }

            // Set the data type as text in the http response object that will
            // be returned.
            result.setIsTextResult(true);

            // Set the text data in the http response object that will
            // be returned.
            if (response.length() > 0) {
                result.setTextResult(response.toString());
            }

            // Set the response code and response message in the http response
            // object that will be returned.
            int rcode = httpUC.getResponseCode();
            this.responseCode = rcode; // QAI 90061 -- Digest Authentication
            String rmsg = httpUC.getResponseMessage();
            result.setResponseCode(rcode);
            result.setResponseMessage(rmsg);
            if (this.logger.isDebugEnabled() == true) {
                this.logger.debug("Got response code [" + rcode + "]");
                this.logger.debug("Got response message [" + rmsg + "]");
            }

            // If content returned contains <META HTTP-EQUIV="Refresh" ....>
            // Then set redirected URL and start post or get over
            String metaRedirectURL = checkAndGetMetaRedirectURL(
                    response.toString());

            if (metaRedirectURL != null) {
                this.logger.info("Got meta data redirect message; redirected URL [" + metaRedirectURL + "]");

                // But first, save the original URL
                this.savedURLOnRedirect = this.urlString;
                this.urlString = metaRedirectURL;

                // Start post or get again with new https url
                try {
                    if (posting) {
                        if (this.logger.isDebugEnabled() == true) {
                            this.logger.debug("Re-POST with URL [" + urlString + "]");
                        }
                        return this.post();
                    } else {
                        if (this.logger.isDebugEnabled() == true) {
                            this.logger.debug("Re-GET with URL [" + urlString + "]");
                        }
                        return this.get();
                    }
                } catch (Exception ex) {
                    throw new IOException(ex.toString());
                }
            }
        }

        return result;
    }

    /**
     * Attempt to read the message body from the error stream - in cases where status code returned
     * is anything other than 200s.  Assumes error text returned!
     *
     * @param httpC HttpURLConnection
     * @param result HttpResult
     * @return boolean
     */
    private boolean getFromErrorStream (HttpResult result) {
        boolean success = true;
        char[] dataChunk = new char[BUFF_SIZE];
        BufferedReader in = null;

        // Get the error stream
        try {
            if ( (encoding.length() == 0) || encoding.equals("ASCII")) {
                in = new BufferedReader(
                    new InputStreamReader(httpUC.getErrorStream()));
            }
            else {
                in = new BufferedReader(
                    new InputStreamReader(
                    httpUC.getErrorStream(),
                    encoding));
            }
        } catch (Throwable t) {
            logger.error("Error occurred while trying to get the error stream from the connection.", t);
            success = true;
        }

        // Read from the error stream for the message.
        if (in != null) {
            try {
                StringBuffer response = new StringBuffer();
                int dataRead = in.read(dataChunk);

                while (dataRead != -1) {
                    response.append(
                        dataChunk,
                        0,
                        dataRead);
                    dataRead = in.read(dataChunk);
                }
                if (this.logger.isDebugEnabled() == true) {
                    this.logger.debug("Got response from error stream [" + response.toString() +
                                      "]");
                }
                in.close();

                result.setTextResult(response.toString());
            }
            catch (Throwable t) {
                logger.error(
                    "Error occurred while reading from the error stream.", t);
                success = false;
            }
        }

        try {
            // Set the response code and response message in the http response
            // object that will be returned.
            int rcode = httpUC.getResponseCode();
            this.responseCode = rcode; // QAI 90061 -- Digest Authentication
            String rmsg = httpUC.getResponseMessage();
            result.setResponseCode(rcode);
            result.setResponseMessage(rmsg);
            if (this.logger.isDebugEnabled() == true) {
                this.logger.debug("Got response code [" + rcode + "]");
                this.logger.debug("Got response message [" + rmsg + "]");
            }
        } catch (Throwable t) {
            logger.error("Error occurred while getting the response code and response message", t);
            success = false;
        }

        return success;
    }

    private String checkAndGetMetaRedirectURL(String httpBodyResponse) {
        // A Meta http-equiv tag looks like this :
        // <META HTTP-EQUIV="Refresh"  CONTENT="0; URL=http://www.ebay.com/">
        // or
        // <meta http-equiv="refresh" content="5;
        //  URL=http://www.wiu.edu/users/mibus/billrec.htm">
        //
        // First Lowercase all
        String temp = new String(httpBodyResponse);
        temp = temp.toLowerCase();

        // Check if contain Meta tag, otherwise return
        int metaTagIndex = temp.indexOf("<meta");

        if (metaTagIndex == -1) {
            return null;
        } else {
            // Check if http-equiv="refresh" after meta tag, otherwise return
            if (temp.indexOf(
                        "http-equiv=\"refresh\"",
                        metaTagIndex) == -1) {
                return null;
            }
        }

        // Now remove all spaces
        temp = removeSpaces(temp);

        // Then lowercase all
        temp = temp.toLowerCase();

        String urlRedirect = null;

        int metaHttpEquivRefreshIndex = temp.indexOf(
                META_HTTP_EQUIV_REFRESH);

        if (metaHttpEquivRefreshIndex > -1) {
            int urlIndex = temp.indexOf(
                    "url=",
                    metaHttpEquivRefreshIndex);

            if (urlIndex > -1) {
                int quoteIndex = temp.indexOf(
                        '"',
                        urlIndex);
                urlRedirect = new String(
                        temp.substring(
                            urlIndex + 4,
                            quoteIndex));
            }
        }

        return urlRedirect;
    }

    private String removeSpaces(String s) {
        StringTokenizer st = new StringTokenizer(s, " ", false);
        String t = "";

        while (st.hasMoreElements()) {
            t += st.nextElement();
        }

        return t;
    }

    // Checks whether the result from the server is text data.
    private boolean isTextResult(String contentType) {
        boolean isText = false;

        if (contentType.startsWith(CONTENT_TYPE_TEXT)) {
            isText = true;
        } else if (contentType.equalsIgnoreCase(CONTENT_TYPE_URLENCODED)) {
            isText = true;
        }

        return isText;
    }

    private boolean isJDK13only() {
        String v = System.getProperty(
                "java.class.version",
                "44.0");
        return ("48.0".compareTo(v) > 0) && ("47.0".compareTo(v) <= 0);
    }

    private boolean needToSetHeaders () {
         if ( (this.proxyUserName != null && this.proxyUserName.length() > 0) ||
              (this.httpUserName != null && this.httpUserName.length() > 0)) {
             return true;
         } else {
             return false;
         }
    }
    
    // QAI 90061 -- Digest Authentication
    //
    // Sample challenge from Tomcat server:
    //
    // Digest realm="private area protected by Digest", nonce="HK78go4CBAA=a13d2d7e7e95f8aeaacd994cff1c17501fedc449", algorithm=MD5, qop="auth"
    //
    private int parse_auth_challenge (String full_challenge) {
      
      int mandatory_count = 0;
      String response_field = "";
      String challenge_response = "";
      
      if (full_challenge != null & full_challenge.length() > 0) {

         boolean digestChallenge = full_challenge.startsWith(DIGEST);
         boolean basicChallenge = full_challenge.startsWith(BASIC);

         //if (pos1 == 0 || pos2 == 0) {
         if (digestChallenge || basicChallenge) {

            if (digestChallenge) {
               challenge_response = full_challenge.substring(DIGEST.length());
            } else if (basicChallenge) {
               challenge_response = full_challenge.substring(BASIC.length());
            }
            challenge_response = challenge_response.trim();
   
            StringTokenizer st = 
               new StringTokenizer(challenge_response, ",", false);
      
            while (st.hasMoreElements()) {
      
               response_field = (String) st.nextElement();

               if (response_field != null && response_field.length() > 0) {
                  mandatory_count += parse_challenge_fields(response_field);

               } else {
                  this.logger.warn("There is an empty or null field/directive in the " + WWW_AUTHENTICATE + " header that will be ignored.");
                  
               } // end "if (response_field != null && response_field.length() > 0) {"
   
            } // end "while (st.hasMoreElements()) {"

            if (digestChallenge && (mandatory_count >= WWW_AUTHENTICATE_DIGEST_MIN_FIELD_COUNT)) {
               if (this.logger.isDebugEnabled() == true) {
                   this.logger.debug("The " + WWW_AUTHENTICATE + " digest header is valid.");
               }
            } else {
               if (basicChallenge && (mandatory_count >= WWW_AUTHENTICATE_BASIC_MIN_FIELD_COUNT)) {
                  if (this.logger.isDebugEnabled() == true) {
                      this.logger.debug("The " + WWW_AUTHENTICATE + " basic header is valid.");
                  }
               } else {
                  this.logger.warn("The " + WWW_AUTHENTICATE + " header is NOT valid because the minimum number " + 
                                    "of mandatory directives had not been detected.");
               }
            }

         } else {
            this.logger.warn("The " + WWW_AUTHENTICATE + " header is NOT valid because the " + 
                              "keyword Basic/Digest is not found in the " + WWW_AUTHENTICATE + " header.");
         }

      } else {         
         this.logger.warn("The " + WWW_AUTHENTICATE + " header is NOT valid because the it is empty or null.");
      }
      // end "if (full_challenge != null & full_challenge.length() > 0) {"

      return mandatory_count;
    }

    private int parse_challenge_fields(String response_field) {

      int pos = 0;
      int mandatory_count = 0;
      String field = response_field.trim();

      if (field.startsWith(NONCE)) {
         /*
         From RFC 2617: http://community.roxen.com/developers/idocs/rfc/rfc2617.html
         ...
         A server-specified data string which should be uniquely generated
         each time a 401 response is made. It is recommended that this
         string be base64 or hexadecimal data. Specifically, since the
         string is passed in the header lines as a quoted string, the
         double-quote character is not allowed.
         The contents of the nonce are implementation dependent.
         ...
         An implementation might choose not to accept a previously used
         nonce or a previously used digest, in order to protect against a
         replay attack. Or, an implementation might choose to use one-time
         nonces or digests for POST or PUT requests and a time-stamp for GET
         requests...The nonce is opaque to the client.
         */
         pos = field.indexOf(NONCE + "=");
         this.digest_auth_nonce = field.substring(pos+(NONCE + "=").length());
         
         if (!this.digest_auth_nonce_last.equals(this.digest_auth_nonce)) {
            // Reset nonce counter if a new nonce has arrived.
            this.nonceCounter = 1;
         }

         this.digest_auth_nonce_last = this.digest_auth_nonce;
         
         mandatory_count++; // nonce is mandatory
         
      } else if (field.startsWith(OPAQUE)) {
         /*
         From RFC 2617: http://community.roxen.com/developers/idocs/rfc/rfc2617.html
         ...
         A string of data, specified by the server, which should be returned
         by the client unchanged in the Authorization header of subsequent
         requests with URIs in the same protection space. It is recommended
         that this string be base64 or hexadecimal data.
         */
         pos = field.indexOf(OPAQUE + "=");
         this.digest_auth_opaque = field.substring(pos+(OPAQUE + "=").length());
         
      } else if (field.startsWith(REALM)) { // Common to both Basic and Digest Authentication

         // TODO: this means we might have to allow user to configure a list/table of
         // user-name-password according to realm but out of requirement scope now;
         // However, with QAI 70636 allowing users to set credentials in collaboration
         // thru OTD methods exposed, it is less of an issue.
         /*
         From RFC 2617: http://community.roxen.com/developers/idocs/rfc/rfc2617.html
         ...
         A string to be displayed to users so they know which username and
         password to use. This string should contain at least the name of
         the host performing the authentication and might additionally
         indicate the collection of users who might have access. An example
         might be "registered_users@gotham.news.com".
         */                  
         pos = field.indexOf(REALM + "=");
         this.digest_auth_realm = field.substring(pos+(REALM + "=").length());
         
         mandatory_count++; // realm is mandatory
         
      } else if (field.startsWith(DOMAIN)) {
         // TODO: Does this mean that we need to add smart logic to determine
         // whether the same authentication info (user name/password) will be
         // used for the underlined url if it is similiar to the domain field?
         // but out of requirement scope now; however, with QAI 70636 allowing 
         // users to set credentials in collaboration thru OTD methods exposed, 
         // it is less of an issue.
         /*
         From RFC 2617: http://community.roxen.com/developers/idocs/rfc/rfc2617.html
         ...
         A quoted, space-separated list of URIs, as specified in RFC XURI, 
         that define the protection space.  If a URI is an abs_path, it
         is relative to the canonical root URL of
         the server being accessed. An absoluteURI in this list may refer to
         a different server than the one being accessed. The client can use
         this list to determine the set of URIs for which the same
         authentication information may be sent: any URI that has a URI in
         this list as a prefix (after both have been made absolute) may be
         assumed to be in the same protection space. If this directive is
         omitted or its value is empty, the client should assume that the
         protection space consists of all URIs on the responding server.
         This directive is not meaningful in Proxy-Authenticate headers, for
         which the protection space is always the entire proxy; if present
         it should be ignored.
         */
         pos = field.indexOf(DOMAIN + "=");
         this.digest_auth_domain = field.substring(pos+(DOMAIN + "=").length());
         
      } else if (field.startsWith(STALE)) {
         // TODO: right now we simply log the stale condition
         /*
         From RFC 2617: http://community.roxen.com/developers/idocs/rfc/rfc2617.html
         ...
         A flag, indicating that the previous request from the client was
         rejected because the nonce value was stale. If stale is TRUE
         (case-insensitive), the client may wish to simply retry the request
         with a new encrypted response, without reprompting the user for a
         new username and password. The server should only set stale to TRUE
         if it receives a request for which the nonce is invalid but with a
         valid digest for that nonce (indicating that the client knows the
         correct username/password). If stale is FALSE, or anything other
         than TRUE, or the stale directive is not present, the username
         and/or password are invalid, and new values must be obtained.
         */
         pos = field.indexOf(STALE + "=");
         this.digest_auth_stale = field.substring(pos+(STALE + "=").length());
         
         // case1: The user and/or password might be wrong to begin with --> the internal re-retry logic
         //        will simply finish and user will not be able to get the page
         // case2: The user and/or password might have been changed on the server side, and once again
         //          the re-retry logic will take care but we might want to warn the user of this case.
         //
         if (!this.digest_auth_stale.equalsIgnoreCase("TRUE")) { // anything other than TRUE

            String msg = "Authentication digest challenge from the server indicating that the nonce " +
                        "value [" + this.digest_auth_nonce + "] from previous request was stale and retry will take place; " + 
                        "please also verify that the user name and/or password is correct and up to date";
            this.logger.warn(msg);
         }
         
      } else if (field.startsWith(ALGORITHM)) {
         // a. My interpretation is that MD5-sess for the algorithm will revert to md5 in digest 
         // algorithm but the calculation will be a bit different
         // a. My interpretation is that "token" for the algorithm means other algorithm other than md5
         /*
         From RFC 2617: http://community.roxen.com/developers/idocs/rfc/rfc2617.html
         ...
         A string indicating a pair of algorithms used to produce the digest
         and a checksum. If this is not present it is assumed to be "MD5".
         If the algorithm is not understood, the challenge should be ignored
         (and a different one used, if there is more than one).
         ... 
         The "MD5-sess" algorithm is intended to
         allow efficient 3rd party authentication servers
         */
         pos = field.indexOf(ALGORITHM + "=");
         this.digest_auth_algorithm = field.substring(pos+(ALGORITHM + "=").length());
         
      } else if (field.startsWith(QOP)) {
         /*
         From RFC 2617: http://community.roxen.com/developers/idocs/rfc/rfc2617.html
         ...
         This directive is optional, but is made so only for backward
         compatibility with RFC 2069; it SHOULD be used by all
         implementations compliant with this version of the Digest scheme.
         If present, it is a quoted string of one or more tokens indicating
         the "quality of protection" values supported by the server.  The
         value "auth" indicates authentication; the value "auth-int"
         indicates authentication with integrity protection; see the
         descriptions below for calculating the response directive value for
         the application of this choice. Unrecognized options MUST be
         ignored.
         */
         pos = field.indexOf(QOP + "=");
         this.digest_auth_qop = field.substring(pos+(QOP + "=").length());
         
         if (this.digest_auth_qop != null && this.digest_auth_qop.length() > 0) {

            StringTokenizer tok = 
               new StringTokenizer(trim_quotes(this.digest_auth_qop), ",");

            while (tok.hasMoreTokens()) {

               String qop = tok.nextToken().trim();

               if (qop.equals("auth")) {
                  this.digest_auth_qop_filtered = qop;
                  break; // The first one obtained is presumed to be the preferred qop

               } else if (qop.equals("auth-int")) {
                  this.digest_auth_qop_filtered = qop;
                  break; // The first one obtained is presumed to be the preferred qop

               } else {
                  this.digest_auth_qop_filtered = "";
                  this.logger.warn("Unsupported qop detected: "+ qop);

               }// end "if (qop.equals("auth")) {"
     
            } // end 'while (tok.hasMoreTokens()) {"

            this.digest_auth_qop = this.digest_auth_qop_filtered;

         } // end "if (this.digest_auth_qop != null && this.digest_auth_qop.length() > 0) {"

      } else {
         this.logger.debug("The directive and field value " + response_field + 
                           " in the " + WWW_AUTHENTICATE + " header is not supported.");

      } // end "if (response_field.startsWith(NONCE)) {"
      
      return mandatory_count;
    }

    // QAI 90061 -- Digest Authentication
    //
    private String trim_quotes(String quoted_sting) {
      
      String rtn = quoted_sting;

      if (quoted_sting != null && quoted_sting.length() > 0) {
         if (quoted_sting.startsWith("\"") &&
             quoted_sting.endsWith("\"") ) {
   
            rtn = quoted_sting.substring(1, quoted_sting.length() - 1);
            
         } else if (quoted_sting.startsWith("\"") &&
                    !quoted_sting.endsWith("\"") ) {

            rtn = quoted_sting.substring(1, quoted_sting.length());

         } else if (!quoted_sting.startsWith("\"") &&
                    quoted_sting.endsWith("\"") ) {

            rtn = quoted_sting.substring(0, quoted_sting.length() - 1);
         }

      }
      //else {} // Don't care, just return the original input.

      return rtn;
    }

    // QAI 90061 -- Digest Authentication
    //
    private String getDigestAuthorizationCredentials(String httpUserName,
                                                   String httpPassWord) throws Exception {

      String nonceCount = "";
      String method = "GET"; // Default for now
      String uri_for_proxy = this.urlString;
      String response, HA1, HA2 = "";
      String credentials = "";      
      String cnonce = CLIENT_NONCE;
      String hexString = Integer.toHexString(this.nonceCounter);

      if (hexString != null && hexString.length() > 0) {
         StringBuffer hexStringBuf = new StringBuffer(hexString);
         while (hexStringBuf.length() < 8) {
            // Pad with leading zero(s) to make it 8LHEX per RFC 2617
            hexStringBuf.insert(0, "0"); 
         }      
         nonceCount = hexStringBuf.toString();
      }

      // Method is part of response calculation but 
      // not for the header      
      if (posting) {
         method = "POST";
      } else {
         method = "GET";
      }
         
      HTTPDigest httpDigest = new HTTPDigest(this.digest_auth_algorithm);

      try {
         HA1 = 
            httpDigest.DigestCalcHA1(this.digest_auth_algorithm, 
                                    httpUserName, 
                                    trim_quotes(this.digest_auth_realm),
                                    httpPassWord, 
                                    trim_quotes(this.digest_auth_nonce), 
                                    cnonce);
         response = 
            httpDigest.DigestCalcResponse(HA1, 
                                          trim_quotes(this.digest_auth_nonce), 
                                          nonceCount, 
                                          cnonce, 
                                          trim_quotes(this.digest_auth_qop), 
                                          method, 
                                          uri_for_proxy, 
                                          HA2,
                                          this.digest_auth_algorithm);
      
         credentials = DIGEST + " " +

                        //The user's name in the specified realm.
                        USERNAME + "=\"" + httpUserName + "\"," +
                        ((this.digest_auth_realm.length() <= 0) ? "": REALM + "=" + this.digest_auth_realm + ",") +

                        // TODO: handle Authenticate-Info from web server for verifying the 
                        // server (not a requirement as of now)
                        /*
                        From RFC 2617: http://community.roxen.com/developers/idocs/rfc/rfc2617.html
                         ...
                        I guess the nonce must be those supplied in the 
                        WWW-Authenticate response header or Authenticate-Info header.
                        */
                        ((this.digest_auth_nonce.length() <= 0) ? "": NONCE + "=" + this.digest_auth_nonce + ",") +

                        /*
                        From RFC 2617: http://community.roxen.com/developers/idocs/rfc/rfc2617.html
                         ...
                        This MUST be specified if a qop directive is sent, and
                        MUST NOT be specified if the server did not send a qop directive in
                        the WWW-Authenticate header field.  The cnonce-value is an opaque
                        quoted string value provided by the client and used by both client
                        and server to avoid chosen plaintext attacks, to provide mutual
                        authentication, and to provide some message integrity protection.
                        See the descriptions below of the calculation of the response-
                        digest and request-digest values.
                        */
                        (((this.digest_auth_qop.length() <= 0) 
                           || (cnonce.length() <= 0)) ? "": CNONCE + "=\"" + cnonce + "\",") +

                        /*
                        From RFC 2617: http://community.roxen.com/developers/idocs/rfc/rfc2617.html
                         ...
                        This MUST be specified if a qop directive is sent (see above), and
                        MUST NOT be specified if the server did not send a qop directive in
                        the WWW-Authenticate header field.  The nc-value is the hexadecimal
                        count of the number of requests (including the current request)
                        that the client has sent with the nonce value in this request.  For
                        example, in the first request sent in response to a given nonce
                        value, the client sends "nc=00000001".  The purpose of this
                        directive is to allow the server to detect request replays by
                        maintaining its own copy of this count - if the same nc-value is
                        seen twice, then the request is a replay.
                        */
                        (((this.digest_auth_qop.length() <= 0) || 
                           ((nonceCount != null) && (nonceCount.length() <= 0))) ? "": NONCECOUNT + "=" + nonceCount + ",") +

                        /*
                        From RFC 2617: http://community.roxen.com/developers/idocs/rfc/rfc2617.html
                         ...
                        Indicates what "quality of protection" the client has applied to
                        the message. If present, its value MUST be one of the alternatives (i.e. auth,auth-int)
                        the server indicated it supports in the WWW-Authenticate header.
                        These values affect the computation of the request-digest. Note
                        that this is a single token, not a quoted list of alternatives as
                        in WWW-Authenticate.  This directive is optional in order to
                        preserve backward compatibility with a minimal implementation of
                        RFC 2069, but SHOULD be used if the server indicated that qop
                        is supported by providing a qop directive in the WWW-Authenticate
                        header field
                        */
                        ((this.digest_auth_qop.length() <= 0) ? "": "qop=" + trim_quotes(this.digest_auth_qop) + ",") +
                        
                        /*
                        From RFC 2617: http://community.roxen.com/developers/idocs/rfc/rfc2617.html
                         ...
                        The URI from Request-URI of the Request-Line; duplicated here
                        because proxies are allowed to change the Request-Line in transit.
                        ...
                        The purpose of duplicating information from the request URL in this 
                        field is to deal with the possibility that an intermediate proxy may
                        alter the client's Request-Line. This altered (but presumably semantically 
                        equivalent) request would not result in the same digest as that calculated
                        by the client. 
                        ...
                        */
                        // QAI 95669: indirect consequence of QAI 93942
                        // URI could be empty
                        ((uri_for_proxy == null || uri_for_proxy.length() <= 0) ? "": URI + "=\"" + uri_for_proxy + "\",") + 

                        /*
                        From RFC 2617: http://community.roxen.com/developers/idocs/rfc/rfc2617.html
                         ...
                        The values of the opaque and algorithm fields must be those supplied in the 
                        WWW-Authenticate response header for the entity being requested.
                        */
                        ((this.digest_auth_opaque.length() <= 0) ? "": OPAQUE + "=" + this.digest_auth_opaque + ",") +
                        ((this.digest_auth_algorithm.length() <= 0) ? "": ALGORITHM + "=" + this.digest_auth_algorithm + ",") +

                        "response=\"" + response + "\"";
                        
         nonceCounter = nonceCounter + 1; // Bump the count if the server nonce is to be reused.
      
      } catch (Exception e) {
         this.logger.error("getDigestAuthorizationCredentials: exception caught: " + e.getMessage());
         // TODO: might want to make stop on error (with exception) configurable;
         // some user might not care but some might do
         //throw new IOException(e.getMessage()); 

      } // end "try {"

      return credentials;
   }

   // QAI 90061 -- Digest Authentication
   // QAI 70636
   //
   // No corresponding getter (i.e. getCredentials) for obvious
   // security precaution.
   //
   public void setCredentials(String userName, String password) {
      setHttpUserName(userName);
      setHttpPassWord(password);
   }

   // QAI 90061 -- Digest Authentication
   //  
   // User can set the strongest authentication scheme level
   //
   public void setStrongestAuthSchemeLevel(int schemeCode) {
      this.authScheme = schemeCode;
   }
   // QAI 90061 -- Digest Authentication
   //
   // User can get the current strongest authentication scheme level
   //
   public int getStrongestAuthSchemeLevel() {
      return this.authScheme;
   }
}