/**************************************************************************
 *
 *          Copyright (c) 2002, SeeBeyond Technology Corporation,
 *          All Rights Reserved
 *
 *          This program, and all the routines referenced herein,
 *          are the proprietary properties and trade secrets of
 *          SEEBEYOND TECHNOLOGY CORPORATION.
 *
 *          Except as provided for by license agreement, this
 *          program shall not be duplicated, used, or disclosed
 *          without  written consent signed by an officer of
 *          SEEBEYOND TECHNOLOGY CORPORATION.
 *
 ***************************************************************************/
package com.stc.connector.httpadapter.http;

import java.security.Security;
import java.security.Provider;


/**
 * Wrapper class for Security related SSL properties.
 *
 * @author James D. Tran
 * @version $Revision: 1.2 $
 */
public final class HttpsSecurityProperties {
    /** X509 Certificate implementation property */
    protected static final String X509_CERTIFICATE_IMPL_PROPERTY = "cert.provider.x509v1";
    /** SSL SocketFactory implementation property */
    protected static final String DEFAULT_SSL_SOCKETFACTORY_IMPL_PROPERTY = "ssl.SocketFactory.provider";
    /** SSL ServerSocketFactory implementation property */
    protected static final String DEFAULT_SSL_SSLSERVERSOCKETFACTORY_IMPL_PROPERTY =
        "ssl.ServerSocketFactory.provider";
    /** Default KeyManager algorithm name property */
    protected static final String DEFAULT_KEYMANAGER_ALGORITHM_NAME = "sun.ssl.keymanager.type";
    /** Default TrustManager algorithm name property */
    protected static final String DEFAULT_TRUSTMANAGER_ALGORITHM_NAME = "sun.ssl.trustmanager.type";

    /**
     * Add a Cryptographic Service Provider (provider).   This method will add
     * a JSSE  provider implementation to the list of provider
     * implementations. This method must  be called, after calling
     * HttpsSystemProperties.setHttpsImplementation, in order to use HTTPs.
     *
     * @param providerClass The JSSE Crytographic Service Provider to add to
     *        the list of JSSE  provider implementations. For example, to use
     *        SUN's JSSE Provider reference implementation,  specify
     *        "com.sun.net.ssl.internal.ssl.Provider".
     *
     * @throws Exception upone error.
     */
    public static void addProvider(String providerClass)
        throws Exception {
        Provider provider = null;

        try {
            Class clsFactory = Class.forName(providerClass);
            provider = (Provider) clsFactory.newInstance();
        } catch (Exception ex) {
            throw new Exception("Failed to create provider class "
                + providerClass + "; Exception : " + ex.toString());
        }

        try {
            Security.addProvider(provider);
        } catch (Exception ex) {
            throw new Exception("Failed to add provider " + providerClass
                + "; Exception : " + ex.toString());
        }
    }

    /**
     * Add a Cryptographic Service Provider (provider).   This method will add
     * a JSSE  provider implementation to the list of provider
     * implementations. This method must  be called, after calling
     * HttpsSystemProperties.setHttpsImplementation, in order to use HTTPs.
     * The Provider will be inserted into the list at the position  indicated
     * by position.
     *
     * @param providerClass The JSSE Crytographic Service Provider to add to
     *        the list of JSSE  provider implementations. For example, to use
     *        SUN's JSSE Provider reference implementation,  specify
     *        "com.sun.net.ssl.internal.ssl.Provider".
     * @param position The position to insert this Provider.
     *
     * @throws Exception upon error.
     */
    public static void insertProviderAt(String providerClass, int position)
        throws Exception {
        Provider provider = null;

        try {
            Class clsFactory = Class.forName(providerClass);
            provider = (Provider) clsFactory.newInstance();
        } catch (Exception ex) {
            throw new Exception("Failed to create provider class "
                + providerClass + "; Exception : " + ex.toString());
        }

        try {
            Security.insertProviderAt(
                provider,
                position);
        } catch (Exception ex) {
            throw new Exception("Failed to insert provider  " + providerClass
                + " at position " + position + "; Exception : " + ex.toString());
        }
    }

    /**
     * Gets a list of Cryptographic Service Provider (provider).  This method
     * will get a list of the providers that were previously set.
     *
     * @return An array of Provider.
     *
     * @see #addProvider
     * @see #insertProviderAt
     * @see java.security.Provider
     */
    public static Provider[] getProviders() {
        return Security.getProviders();
    }

    /**
     * Sets the X509Certificate implementation.
     *
     * @param x509CertificateImpl The implementation class of X509Certificate.
     *        For example, if the implementation class is called
     *        MyX509CertificateImpl and it appears in the  com.radcrypto
     *        package, you should specify com.radcrypto.MyX509CertificateImpl.
     *
     * @throws Exception if unable to set the X509Certificate implementation.
     */
    public static void setX509CertificateImpl(String x509CertificateImpl)
        throws Exception {
        try {
            Security.setProperty(
                X509_CERTIFICATE_IMPL_PROPERTY,
                x509CertificateImpl);
        } catch (Exception ex) {
            throw new Exception(
                "Failed to set X509Certificate implmentation; Exception : "
                + ex.toString());
        }
    }

    /**
     * Gets the X509Certificate implementation.
     *
     * @return The implementation class of X509Certificate. For example, if the
     *         implementation class was set to  MyX509CertificateImpl and it
     *         appears in the com.radcrypto package,  it will return
     *         com.radcrypto.MyX509CertificateImpl.  Null will be returned if
     *         X509Certificate implementation was not set.
     */
    public static String getX509CertificateImpl() {
        return Security.getProperty(X509_CERTIFICATE_IMPL_PROPERTY);
    }

    /**
     * Sets the default SSL Socket Factory implementation.
     *
     * @param sslSocketFactoryImplClass The implementation class of SSL Socket
     *        Factory. For example, if the implementation class is called
     *        MySSLSocketFactoryImpl and it appears in the  com.radcrypto
     *        package, you should specify
     *        com.radcrypto.MySSLSocketFactoryImpl.
     *
     * @throws Exception if unable to set the default SSL Socket Factory
     *         implementation.
     */
    public static void setSSLSocketFactoryImpl(String sslSocketFactoryImplClass)
        throws Exception {
        try {
            Security.setProperty(
                DEFAULT_SSL_SOCKETFACTORY_IMPL_PROPERTY,
                sslSocketFactoryImplClass);
        } catch (Exception ex) {
            throw new Exception(
                "Failed to set the default SSL Socket Factory implmentation; Exception : "
                + ex.toString());
        }
    }

    /**
     * Gets the default SSL Socket Factory implementation.
     *
     * @return The implementation class of SSL Socket Factory. For example, if
     *         the implementation class was set to MySSLSocketFactoryImpl and
     *         it appears in the  com.radcrypto package, it will return
     *         com.radcrypto.MySSLSocketFactoryImpl. If SSL Socket Factory was
     *         not set, it will return null.
     */
    public static String getSSLSocketFactoryImpl() {
        return Security.getProperty(DEFAULT_SSL_SOCKETFACTORY_IMPL_PROPERTY);
    }

    /**
     * Sets the default SSL Server Socket Factory implementation.
     *
     * @param sslServerSocketFactoryImplClass The implementation class of SSL
     *        Server Socket Factory. For example, if the implementation class
     *        is called  MySSLServerSocketFactoryImpl and it appears in the
     *        com.radcrypto package, you should specify
     *        com.radcrypto.MySSLServerSocketFactoryImpl.
     *
     * @throws Exception if unable to set the default SSL Server Socket Factory
     *         implementation.
     */
    public static void setSSLServerSocketFactoryImpl(
        String sslServerSocketFactoryImplClass)
        throws Exception {
        try {
            Security.setProperty(
                DEFAULT_SSL_SSLSERVERSOCKETFACTORY_IMPL_PROPERTY,
                sslServerSocketFactoryImplClass);
        } catch (Exception ex) {
            throw new Exception(
                "Failed to set the default SSL Server Socket Factory implmentation; Exception : "
                + ex.toString());
        }
    }

    /**
     * Gets the default SSL Server Socket Factory implementation.
     *
     * @return The implementation class of SSL Server Socket Factory. For
     *         example, if the implementation class was set to
     *         MySSLServerSocketFactoryImpl and it appears in the
     *         com.radcrypto package, it will return
     *         com.radcrypto.MySSLServerSocketFactoryImpl. If SSL Server
     *         Socket Factory was not set, it will return null.
     */
    public static String getSSLServerSocketFactoryImpl() {
        return Security.getProperty(
            DEFAULT_SSL_SSLSERVERSOCKETFACTORY_IMPL_PROPERTY);
    }

    /**
     * Sets the default Key Manager Algorithm name.
     *
     * @param keyManagerAlgoName The name of the key manager algorithm to use.
     *        For example, the default key manager algorithm use in SUN's
     *        reference implementation of JSSE is "SunX509".
     *
     * @throws Exception if unable to set the default key manager algorithm
     *         name.
     */
    public static void setKeyManagerAlgorithm(String keyManagerAlgoName)
        throws Exception {
        try {
            Security.setProperty(
                DEFAULT_KEYMANAGER_ALGORITHM_NAME,
                keyManagerAlgoName);
        } catch (Exception ex) {
            throw new Exception(
                "Failed to set the default Key Manager algorithm name; Exception : "
                + ex.toString());
        }
    }

    /**
     * Gets the default Key Manager Algorithm name previously set.
     *
     * @return The name of the key manager algorithm that was previously set.
     */
    public static String getKeyManagerAlgorithm() {
        return Security.getProperty(DEFAULT_KEYMANAGER_ALGORITHM_NAME);
    }

    /**
     * Sets the default Trust Manager Algorithm name.
     *
     * @param trustManagerAlgoName The name of the trust manager algorithm to
     *        use. For example, the default trust manager algorithm use in
     *        SUN's reference implementation of JSSE is "SunX509".
     *
     * @throws Exception if unable to set the default trust manager algorithm
     *         name.
     */
    public static void setTrustManagerAlgorithm(String trustManagerAlgoName)
        throws Exception {
        try {
            Security.setProperty(
                DEFAULT_TRUSTMANAGER_ALGORITHM_NAME,
                trustManagerAlgoName);
        } catch (Exception ex) {
            throw new Exception(
                "Failed to set the default Trust Manager algorithm name; Exception : "
                + ex.toString());
        }
    }

    /**
     * Gets the default Trust Manager Algorithm name previously set.
     *
     * @return The name of the trust manager algorithm that was previously set.
     */
    public static String getTrustManagerAlgorithm() {
        return Security.getProperty(DEFAULT_TRUSTMANAGER_ALGORITHM_NAME);
    }
}
