/**************************************************************************
 *
 *          Copyright (c) 2002, SeeBeyond Technology Corporation,
 *          All Rights Reserved
 *
 *          This program, and all the routines referenced herein,
 *          are the proprietary properties and trade secrets of
 *          SEEBEYOND TECHNOLOGY CORPORATION.
 *
 *          Except as provided for by license agreement, this
 *          program shall not be duplicated, used, or disclosed
 *          without  written consent signed by an officer of
 *          SEEBEYOND TECHNOLOGY CORPORATION.
 *
 ***************************************************************************/
package com.stc.connector.httpadapter.http;

import sun.misc.BASE64Encoder;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.InetAddress;
import java.net.Socket;
import java.net.UnknownHostException;
import javax.net.ssl.SSLSocket;
import javax.net.ssl.SSLSocketFactory;


/**
 * SSLSocket factory class for proxy tunnel.
 *
 * @author James D. Tran
 * @version $Revision: 1.3 $
 */
public class SSLTunnelSocketFactory
    extends SSLSocketFactory {
    private SSLSocketFactory dfactory;
    private String tunnelHost;
    private int tunnelPort;
    private String username;
    private String password;
    boolean handshakeDone = false;

    /**
     * Constructor for SSLTunnelSocketFactory.
     *
     * @param sslContext The SSLContextJDKPlatform for JDK  specific
     *        SSLContext.
     * @param proxyhost The proxy host.
     * @param proxyport The proxy port.
     * @param proxyusername The proxy username.
     * @param proxypassword The proxy password.
     *
     * @throws IOException upon error.
     */
    public SSLTunnelSocketFactory(SSLContextJDKPlatform sslContext,
        String proxyhost, String proxyport, String proxyusername,
        String proxypassword)
        throws IOException {
        tunnelHost = proxyhost;
        tunnelPort = Integer.parseInt(proxyport);
        username = proxyusername;
        password = proxypassword;

        if (sslContext == null) {
            dfactory = (SSLSocketFactory) javax.net.ssl.SSLSocketFactory
                .getDefault();
        } else {
            dfactory = sslContext.getSocketFactory();
        }
    }

    /**
     * Constructor for SSLTunnelSocketFactory.
     *
     * @param sslContext The SSLContextJDKPlatform for JDK  specific
     *        SSLContext.
     * @param proxyhost The proxy host.
     * @param proxyport The proxy port.
     *
     * @throws IOException upon error.
     */
    public SSLTunnelSocketFactory(SSLContextJDKPlatform sslContext,
        String proxyhost, String proxyport)
        throws IOException {
        tunnelHost = proxyhost;
        tunnelPort = Integer.parseInt(proxyport);
        username = "";
        password = "";

        if (sslContext == null) {
            dfactory = (SSLSocketFactory) javax.net.ssl.SSLSocketFactory
                .getDefault();
        } else {
            dfactory = sslContext.getSocketFactory();
        }
    }

    /**
     * Create a socket.
     *
     * @param host The host.
     * @param port The port.
     *
     * @return An instance of Socket.
     *
     * @throws IOException upon error.
     * @throws UnknownHostException upon error.
     */
    public Socket createSocket(String host, int port)
        throws IOException, UnknownHostException {
        return createSocket(
            null,
            host,
            port,
            true);
    }

    /**
     * Create a socket.
     *
     * @param host The host.
     * @param port The port.
     * @param clientHost InetAddress of client host.
     * @param clientPort port number of client host.
     *
     * @return An instance of Socket.
     *
     * @throws IOException upon error.
     * @throws UnknownHostException upon error.
     */
    public Socket createSocket(String host, int port, InetAddress clientHost,
        int clientPort)
        throws IOException, UnknownHostException {
        return createSocket(
            null,
            host,
            port,
            true);
    }

    /**
     * Create a socket.
     *
     * @param host The InetAddress of host.
     * @param port The port.
     *
     * @return An instance of Socket.
     *
     * @throws IOException upon error.
     * @throws UnknownHostException upon error.
     */
    public Socket createSocket(InetAddress host, int port)
        throws IOException, UnknownHostException {
        return createSocket(
            null,
            host.getHostName(),
            port,
            true);
    }

    /**
     * Create a socket.
     *
     * @param address The InetAddress of host.
     * @param port The port.
     * @param clientAddress InetAddress of client host.
     * @param clientPort port number of client host.
     *
     * @return An instance of Socket.
     *
     * @throws IOException upon error.
     */
    public Socket createSocket(InetAddress address, int port,
        InetAddress clientAddress, int clientPort)
        throws IOException {
        return createSocket(
            null,
            address.getHostName(),
            port,
            true);
    }

    /**
     * Create a socket.
     *
     * @param s The current established socket connection.
     * @param host The host.
     * @param port The port.
     * @param autoClose Automatically close socket connection.
     *
     * @return An instance of Socket.
     *
     * @throws IOException upon error.
     * @throws UnknownHostException upon error.
     */
    public Socket createSocket(Socket s, String host, int port,
        boolean autoClose)
        throws IOException, UnknownHostException {
        Socket tunnel = new Socket(tunnelHost, tunnelPort);

        doTunnelHandshake(
            tunnel,
            host,
            port);

        SSLSocket result = (SSLSocket) dfactory.createSocket(
                tunnel,
                host,
                port,
                autoClose);

        return result;
    }

    /**
     * Execute tunnel handshake with proxy.
     *
     * @param tunnel The Socket connection.
     * @param host The host.
     * @param port The port.
     *
     * @throws IOException upon error.
     */
    private synchronized void doTunnelHandshake(Socket tunnel, String host,
        int port)
        throws IOException {
        OutputStream out = tunnel.getOutputStream();
        String authString = username + ":" + password;
        String msg;

        if (authString.equals(":")) {
            msg = "CONNECT " + host + ":" + port + " HTTP/1.0\n"
                + "User-Agent: "
                + sun.net.www.protocol.http.HttpURLConnection.userAgent
                + "\r\n\r\n";
        } else {
            msg = "CONNECT " + host + ":" + port + " HTTP/1.0\n"
                + "User-Agent: "
                + sun.net.www.protocol.http.HttpURLConnection.userAgent + "\n"
                + "Proxy-Authorization:Basic "
                + (new BASE64Encoder()).encode(authString.getBytes())
                + "\r\n\r\n";
        }

        byte[] b;

        try {
            /*
             * We really do want ASCII7 -- the http protocol doesn't change
             * with locale.
             */
            b = msg.getBytes("ASCII7");
        } catch (UnsupportedEncodingException ignored) {
            /*
             * If ASCII7 isn't there, something serious is wrong, but
             * Paranoia Is Good (tm)
             */
            b = msg.getBytes();
        }

        out.write(b);
        out.flush();

        /*
         * We need to store the reply so we can create a detailed
         * error message to the user.
         */
        byte[] reply = new byte[200];
        int replyLen = 0;
        int newlinesSeen = 0;
        boolean headerDone = false; /* Done on first newline */

        InputStream in = tunnel.getInputStream();
        boolean error = false;

        while (newlinesSeen < 2) {
            int i = in.read();

            if (i < 0) {
                throw new IOException("Unexpected EOF from proxy");
            }

            if (i == '\n') {
                headerDone = true;
                ++newlinesSeen;
            } else if (i != '\r') {
                newlinesSeen = 0;

                if (!headerDone && (replyLen < reply.length)) {
                    reply[replyLen++] = (byte) i;
                }
            }
        }

        /*
         * Converting the byte array to a string is slightly wasteful
         * in the case where the connection was successful, but it's
         * insignificant compared to the network overhead.
         */
        String replyStr;

        try {
            replyStr = new String(
                    reply,
                    0,
                    replyLen,
                    "ASCII7");
        } catch (UnsupportedEncodingException ignored) {
            replyStr = new String(
                    reply,
                    0,
                    replyLen);
        }

        /* Look for 200 connection established */
        if (replyStr.toLowerCase().indexOf("200 connection established") == -1) {
            throw new IOException("Super Unable to tunnel through "
                + tunnelHost + ":" + tunnelPort + ".  Proxy returns \""
                + replyStr + "\"");
        }

        /*else   // Handshake OK - don't do again.
           {
                 handshakeDone = true;
           }*/
        /* tunneling Handshake was successful! */
    }

    /**
     * Get the default cipher suites.
     *
     * @return An array of cipher suite Strings.
     */
    public String[] getDefaultCipherSuites() {
        return dfactory.getDefaultCipherSuites();
    }

    /**
     * Get the supported cipher suites.
     *
     * @return An array of cipher suite Strings.
     */
    public String[] getSupportedCipherSuites() {
        return dfactory.getSupportedCipherSuites();
    }
}
