/**************************************************************************
 *
 *          Copyright (c) 2002, SeeBeyond Technology Corporation,
 *          All Rights Reserved
 *
 *          This program, and all the routines referenced herein,
 *          are the proprietary properties and trade secrets of
 *          SEEBEYOND TECHNOLOGY CORPORATION.
 *
 *          Except as provided for by license agreement, this
 *          program shall not be duplicated, used, or disclosed
 *          without  written consent signed by an officer of
 *          SEEBEYOND TECHNOLOGY CORPORATION.
 *
 ***************************************************************************/
package com.stc.connector.httpadapter.http.jdk13;

import com.sun.net.ssl.KeyManagerFactory;
import com.sun.net.ssl.SSLContext;
import com.sun.net.ssl.TrustManagerFactory;
import java.io.ByteArrayInputStream;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.IOException;
import java.security.KeyStore;
import javax.net.ssl.SSLSocketFactory;

import java.util.Enumeration;
import java.security.cert.X509Certificate;

/**
 * Wrapper class for JDK 1.3 + JSSE extension SSLContext class.
 *
 * @author  $Author: jtran $
 * @version $Revision: 1.5 $
 */
public class SSLContextJDK13
    implements com.stc.connector.httpadapter.http.SSLContextJDKPlatform {
    // TrustManagerFactory algorithm
    private static final String TRUST_MGR_FACTORY_ALGO = "SunX509";

    // KeyManagerFactory algorithm
    private static final String KEY_MGR_FACTORY_ALGO = "SunX509";
    
    // SME related stuff
    private static final String STCJKS = "STCJKS";
    private static final String STCJSSEKS = "STCJSSEKS";
    
    SSLContext sslContext = null;

    /**
     * Creates a new instance of SSLContextJDK13
     */
    public SSLContextJDK13() {
    }

    /**
     * Retrieves the SSLSocketFactory from the established underlying
     * SSLContext.
     *
     * @return An instance of javax.net.SSLSocketFactory.
     *
     * @throws IOException upon error.
     */
    public SSLSocketFactory getSocketFactory()
        throws IOException {
        if (this.sslContext == null) {
            throw new IOException("Can not get SSLSocketFactory; "
                + "null SSLContext");
        }

        return sslContext.getSocketFactory();
    }

    /**
     * Initialize the underlying SSLContext.
     *
     * @param sslProtocol The SSL Protocol to use. Valid are SSL, SSLv2, SSLv3,
     *        TLS, TLSv1
     * @param trustStoreType The truststore type (i.e., JKS).
     * @param trustStore The truststore location.
     * @param trustStorePassword The truststore password.
     * @param keyStoreType The keystore type (i.e., JKS).
     * @param keyStore The keystore location.
     * @param keyStorePassword The keystore password.
     * @param x509Algorithm The X509 algorithm name to use for the
     *        trust and key manager factories.
     *
     * @throws IOException upon error.
     */
    public void init(String sslProtocol, String trustStoreType,
        String trustStore, String trustStorePassword, String keyStoreType,
        String keyStore, String keyStorePassword, String x509Algorithm)
        throws IOException {
        // Get SSLContext instance for the SSL protocol specified
        // (SSL, SSLv2, SSLv3, TLS, TLSv1)

        try {
            sslContext = SSLContext.getInstance(sslProtocol);
        } catch (Exception ex) {
            throw new IOException("Failed to get an instance of "
                + "SSLContext of protocol " + sslProtocol + "; Exception : "
                + ex.toString());
        }

        // Get an instance of the KeyStore (truststore)
        // specified by the type (JKS, STCJKS)
        KeyStore theTrustStore = null;

        try {
            theTrustStore = KeyStore.getInstance(trustStoreType);
        } catch (Exception ex) {
            throw new IOException("Failed to get a KeyStore instance "
                + "of type " + trustStoreType + "; Exception : "
                + ex.toString());
        }

        // Load the truststore
        char[] trustStorePassphrase = null;

        if ((trustStorePassword != null) && (trustStorePassword.length() > 0)) {
            trustStorePassphrase = trustStorePassword.toCharArray();
        }

        try {
            InputStream trustStoreIS = null;
            
            if (trustStoreType.equalsIgnoreCase(STCJSSEKS)) {
                trustStoreIS = new ByteArrayInputStream (trustStore.getBytes());
            } else {
                trustStoreIS = new FileInputStream (trustStore);
            }
            
            theTrustStore.load (trustStoreIS, trustStorePassphrase);
            
        } catch (Exception ex) {
            throw new IOException("Failed to load the truststore " + trustStore
                + " of type " + trustStoreType + "; Exception : "
                + ex.toString());
        }

        // Get an instance of a TrustManagerFactory
        TrustManagerFactory tmf = null;

        try {
            tmf = TrustManagerFactory.getInstance(x509Algorithm);
        } catch (Exception ex) {
            throw new IOException("Failed to get a TrustManagerFactory "
                + "instance for algorithm " + x509Algorithm
                + "; Exception : " + ex.toString());
        }

        // Initialize the TrustManagerFactory
        try {
            tmf.init(theTrustStore);
        } catch (Exception ex) {
            throw new IOException("Failed to initialize the "
                + "TrustManagerFactory instance for algorithm "
                + TRUST_MGR_FACTORY_ALGO + " with truststore " + trustStore
                + "; Exception : " + ex.toString());
        }

        // Load the keystore if specified
        KeyManagerFactory kmf = null;

        if ((keyStore != null) && (keyStore.length() > 0)) {
            // Get an instance of the KeyStore (keystore)
            // specified by the type (JKS, STCJKS)
            KeyStore theKeyStore = null;

            try {
                theKeyStore = KeyStore.getInstance(keyStoreType);
            } catch (Exception ex) {
                throw new IOException("Failed to get a KeyStore "
                    + "instance of type " + keyStoreType + "; Exception : "
                    + ex.toString());
            }

            // Get the keystore password
            char[] keyStorePassphrase = null;

            if ((keyStorePassword != null) && (keyStorePassword.length() > 0)) {
                keyStorePassphrase = keyStorePassword.toCharArray();
            }

            // Load the keystore
            try {
                InputStream keyStoreIS = null;
                if (keyStoreType.equalsIgnoreCase(STCJSSEKS)) {
                    keyStoreIS = new ByteArrayInputStream (keyStore.getBytes());
                } else {
                    keyStoreIS = new FileInputStream(keyStore);
                }
                
                theKeyStore.load (keyStoreIS, keyStorePassphrase);
            } catch (Exception ex) {
                throw new IOException("Failed to load the keystore " + keyStore
                    + " of type " + keyStoreType + "; Exception : "
                    + ex.toString());
            }

            // Get a KeyManagerFactory instance
            try {
                kmf = KeyManagerFactory.getInstance(x509Algorithm);
            } catch (Exception ex) {
                throw new IOException("Failed to get a "
                    + "KeyManagerFactory instance for algorithm "
                    + x509Algorithm + "; Exception : " + ex.toString());
            }

            // Initialize the KeyManagerFactory instance with the
            // keystore
            try {
                if (keyStoreType.equalsIgnoreCase(STCJSSEKS)) {
                    kmf.init(
                        theKeyStore,
                        null);
                } else {
                    kmf.init(
                        theKeyStore,
                        keyStorePassphrase);
                }
            } catch (Exception ex) {
                throw new IOException("Failed to initialize the "
                    + "KeyManagerFactory instance for algorithm "
                    + KEY_MGR_FACTORY_ALGO + " with keystore " + keyStore
                    + "; Exception : " + ex.toString());
            }
        }

        // Initialize SSLContext instance with the trustore and keystore
        // from TrustManagerFactory and KeyManagerFactory.
        try {
            sslContext.init(
                (kmf == null)
                ? null : kmf.getKeyManagers(),
                (tmf == null)
                ? null : tmf.getTrustManagers(),
                null);
        } catch (Exception ex) {
            throw new IOException("Failed to initialize the "
                + "SSLContext instance" + "; Exception : " + ex.toString());
        }
    }
}
