/**************************************************************************
 *
 *          Copyright (c) 2002, SeeBeyond Technology Corporation,
 *          All Rights Reserved
 *
 *          This program, and all the routines referenced herein,
 *          are the proprietary properties and trade secrets of
 *          SEEBEYOND TECHNOLOGY CORPORATION.
 *
 *          Except as provided for by license agreement, this
 *          program shall not be duplicated, used, or disclosed
 *          without  written consent signed by an officer of
 *          SEEBEYOND TECHNOLOGY CORPORATION.
 *
 ***************************************************************************/

package com.stc.connector.httpadapter.http.jdk14;

import com.stc.connector.httpadapter.http.SSLContextJDKPlatform;
import com.stc.connector.httpadapter.http.SSLTunnelSocketFactory;
import java.io.IOException;
import java.net.HttpURLConnection;
import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.SSLSession;

/**
 * Wrapper class for SSL/JSSE JDK 1.4 specific methods.
 *
 * @author  $Author: jtran $
 * @version $Revision: 1.4 $
 */
public class HttpClientAPIJDK14
    implements com.stc.connector.httpadapter.http.HttpClientAPIJDKPlatform {
                
    /**
     * Constructor for HttpClientAPIJDK14
     */
    public HttpClientAPIJDK14() {
    }

    /**
     * Set the SSLSocketFactory for the current HttpURLConnection using the
     * SSLContext provided if the connection is an HTTPS connection.
     *
     * @param httpUC The HttpURLConnection for the HTTPS connection.
     * @param sslContext The SSLContextJDKPlatform which provides key
     *        management SSLSocketFactory.
     *
     * @return The boolean true if the connection is an HTTPS connection and
     *         the SSLSocketFactory was set for the connection using the
     *         SSLSocketFactory provided by the SSLContext.
     *
     * @throws IOException upon error.
     */
    public boolean setSocketFactory(HttpURLConnection httpUC,
        SSLContextJDKPlatform sslContext)
        throws IOException {
        boolean sslSockFactSet = false;

        if (httpUC instanceof javax.net.ssl.HttpsURLConnection) {
            ((javax.net.ssl.HttpsURLConnection) httpUC).setSSLSocketFactory(
                sslContext.getSocketFactory());
            sslSockFactSet = true;
        }

        return sslSockFactSet;
    }

    /**
     * Sets the connection to not do hostname verification on the
     * X509 certificate if the connection is an https connection.
     *
     * @param urlConn The HttpURLConnection for the HTTP connection.
     *
     * @return True if connection is https connection and no hostname
     *         verifier was set successfully; otherwise false.
     * 
     * @throws Exception upon error.
     */
     public boolean setNoHostnameVerifier(HttpURLConnection httpUC)
        throws Exception {
        boolean retVal = false;
        if (httpUC instanceof javax.net.ssl.HttpsURLConnection) {
            try {
                ((javax.net.ssl.HttpsURLConnection) httpUC)
                .setHostnameVerifier(new NoHostnameVerifier());
                retVal = true;
            } catch (Exception ex) {
                throw new Exception("Problem setting the no hostname verifier " +
                                    "on the https connection.");
            }
        }
        return retVal;
    }

    /**
     * Set the SSLTunnelSocketFactory for the current HttpURLConnection using
     * the SSLContext provided if the connection is an HTTPS connection.
     *
     * @param httpUC The HttpURLConnection for the HTTPS connection.
     * @param sslContext The SSLContextJDKPlatform which provides key
     *        management SSLSocketFactory.
     * @param host The proxy host name.
     * @param port The proxy port number.
     * @param proxyUserName The proxy username.
     * @param proxyPassWord The proxy password.
     *
     * @return The boolean true if the connection is an HTTPS connection and
     *         the SSLTunnelSocketFactory was set for the connection using the
     *         SSLSocketFactory provided by the SSLContext.
     *
     * @throws IOException upon error.
     */
    public boolean setTunnelSocketFactory(HttpURLConnection httpUC,
        SSLContextJDKPlatform sslContext, String host, int port,
        String proxyUserName, String proxyPassWord)
        throws IOException {
        boolean sslSockFactSet = false;

        if ((host != null) && (port != -1)) {
            if (httpUC instanceof javax.net.ssl.HttpsURLConnection) {
                ((javax.net.ssl.HttpsURLConnection) httpUC).setSSLSocketFactory(
                    new SSLTunnelSocketFactory(
                        sslContext,
                        host,
                        Integer.toString(port),
                        proxyUserName,
                        proxyPassWord));
                sslSockFactSet = true;
            }
        }

        return sslSockFactSet;
    }
    
    // NoHostNameVerifier
    private class NoHostnameVerifier
        implements HostnameVerifier {

        /** 
         * Supply a policy for handling the case where the host 
         * to connect to and the server name from the certificate mismatch.
         * In this case, always returns true so that hostname verification 
         * is essentially disabled.
         * 
         * @param   urlHostname     The URL hostname.
         * @param   session         SSL session hostname is verified in.
         *
         * @return  A boolean true if host verification succeeds; false otherwise.
         */
        public boolean verify(String urlHostname, SSLSession session) {
            return true;
        }
    }            
}
